import copy
import logging
import math
from collections import OrderedDict
from typing import List

import psutil

try:
    import GPUtil
except ImportError:
    raise ImportError(" ray is not installed. In order to use auto_train please run pip install ludwig[ray]")

from ludwig.api import LudwigModel
from ludwig.backend import initialize_backend
from ludwig.constants import (
    AUTO,
    AUTOML_DEFAULT_TEXT_ENCODER,
    AUTOML_LARGE_TEXT_DATASET,
    AUTOML_MAX_ROWS_PER_CHECKPOINT,
    AUTOML_SMALLER_TEXT_ENCODER,
    AUTOML_SMALLER_TEXT_LENGTH,
    AUTOML_TEXT_ENCODER_MAX_TOKEN_LEN,
    HYPEROPT,
    PREPROCESSING,
    SPACE,
    TEXT,
    TRAINER,
)
from ludwig.data.preprocessing import preprocess_for_training
from ludwig.features.feature_registries import update_config_with_metadata
from ludwig.schema.model_config import ModelConfig
from ludwig.utils.automl.utils import get_model_type
from ludwig.utils.torch_utils import initialize_pytorch

logger = logging.getLogger(__name__)

# maps variable search space that can be modified to minimum permissible value for the range
RANKED_MODIFIABLE_PARAM_LIST = {
    "tabnet": OrderedDict(
        {
            "trainer.batch_size": 32,
            "combiner.size": 8,
            "combiner.output_size": 8,
        }
    ),
    "concat": OrderedDict(
        {
            "trainer.batch_size": 32,
            "combiner.output_size": 64,
            "combiner.num_fc_layers": 1,
        }
    ),
    "tabtransformer": OrderedDict(
        {
            "trainer.batch_size": 32,
            "combiner.num_heads:": 4,
            "combiner.output_size": 8,
            "combiner.num_layers": 4,
            "combiner.num_fc_layers": 1,
        }
    ),
    "text": OrderedDict(  # for single input feature text models e.g. bert and its variants
        {
            "trainer.batch_size": 16,
        }
    ),
}


BYTES_PER_MiB = 1048576
BYTES_PER_WEIGHT = 4  # assumes 32-bit precision = 4 bytes
BYTES_OPTIMIZER_PER_WEIGHT = 8  # for optimizer m and v vectors


def get_trainingset_metadata(config, dataset, backend):
    (_, _, _, training_set_metadata) = preprocess_for_training(
        config, dataset=dataset, preprocessing_params=config[PREPROCESSING], backend=backend
    )
    return training_set_metadata


# Note: if run in Ray Cluster, this method is run remote with gpu resources requested if available
def _get_machine_memory():
    if GPUtil.getGPUs():
        machine_mem = GPUtil.getGPUs()[0].memoryTotal * BYTES_PER_MiB
    else:
        machine_mem = psutil.virtual_memory().total
    return machine_mem


def _get_text_feature_max_length(config, training_set_metadata) -> int:
    """Returns max sequence length over text features, subject to preprocessing limit."""
    max_length = 0
    for feature in config["input_features"]:
        if feature["type"] == TEXT:
            feature_max_len = training_set_metadata[feature["name"]]["max_sequence_length"]
            if feature_max_len > max_length:
                max_length = feature_max_len
    if (
        ("preprocessing" in config)
        and (TEXT in config["preprocessing"])
        and ("max_sequence_length" in config["preprocessing"][TEXT])
    ):
        limit = config["preprocessing"][TEXT]["max_sequence_length"]
    else:
        limit = 256  # Preprocessing default max_sequence_length = 256
    if max_length > limit + 2:  # For start and stop symbols.
        max_length = limit + 2
    return max_length


def _get_text_model_memory_usage(config, training_set_metadata, memory_usage) -> int:
    max_feature_token_length = _get_text_feature_max_length(config, training_set_metadata)
    memory_usage = (memory_usage / AUTOML_TEXT_ENCODER_MAX_TOKEN_LEN) * max_feature_token_length
    return memory_usage


def compute_memory_usage(config_obj, training_set_metadata, model_category) -> int:
    update_config_with_metadata(config_obj, training_set_metadata)
    lm = LudwigModel.create_model(config_obj)
    model_size = lm.get_model_size()  # number of parameters in model
    batch_size = config_obj.trainer.batch_size
    if batch_size == AUTO:
        # Smallest valid batch size that will allow training to complete
        batch_size = 2
    memory_usage = model_size * (BYTES_PER_WEIGHT + BYTES_OPTIMIZER_PER_WEIGHT) * batch_size
    if model_category == TEXT:
        return _get_text_model_memory_usage(config_obj.to_dict(), training_set_metadata, memory_usage)
    else:
        return memory_usage


def sub_new_params(config: dict, new_param_vals: dict):
    new_config = copy.deepcopy(config)
    for param, val in new_param_vals.items():
        config_section = param.split(".")[0]
        param_name = param.split(".")[1]
        new_config[config_section][param_name] = val
    return new_config


def get_new_params(current_param_values, hyperparam_search_space, params_to_modify):
    for param, _ in params_to_modify.items():
        if param in hyperparam_search_space:
            if hyperparam_search_space[param][SPACE] == "choice":
                current_param_values[param] = hyperparam_search_space[param]["categories"][-1]
            else:
                current_param_values[param] = hyperparam_search_space[param]["upper"]
    return current_param_values


def _update_text_encoder(input_features: List, old_text_encoder: str, new_text_encoder: str) -> None:
    for feature in input_features:
        if feature["type"] == TEXT and feature["encoder"] == old_text_encoder:
            feature["encoder"] = new_text_encoder


def _get_text_feature_min_usable_length(input_features: List, training_set_metadata) -> int:
    """Returns min of AUTOML_SMALLER_TEXT_LENGTH and lowest 99th percentile sequence length over text features."""
    min_usable_length = AUTOML_SMALLER_TEXT_LENGTH
    for feature in input_features:
        if feature["type"] == TEXT:
            feature_99ptile_len = training_set_metadata[feature["name"]]["max_sequence_length_99ptile"]
            if feature_99ptile_len < min_usable_length:
                min_usable_length = feature_99ptile_len
    return round(min_usable_length)


def reduce_text_feature_max_length(config, training_set_metadata) -> bool:
    """Reduce max sequence length, when viable, to control its quadratic impact."""
    input_features = config["input_features"]
    min_usable_length = _get_text_feature_min_usable_length(input_features, training_set_metadata)
    seq_len_limit = {"max_sequence_length": min_usable_length}
    if "preprocessing" not in config:
        config["preprocessing"] = {TEXT: seq_len_limit}
    elif (
        (TEXT not in config["preprocessing"])
        or ("max_sequence_length" not in config["preprocessing"][TEXT])
        or (min_usable_length < float(config["preprocessing"][TEXT]["max_sequence_length"]))
    ):
        config["preprocessing"][TEXT] = seq_len_limit
    else:
        return False
    return True


# For hyperparam_search_space comprised solely of choice spaces, compute maximum number of
# combinations and return that value if it is less than num_samples; else return num_samples.
def _update_num_samples(num_samples, hyperparam_search_space):
    max_num_samples = 1
    for param in hyperparam_search_space.keys():
        if hyperparam_search_space[param][SPACE] == "choice":
            max_num_samples *= len(hyperparam_search_space[param]["categories"])
        else:
            return num_samples
    if max_num_samples < num_samples:
        return max_num_samples
    return num_samples


# Note: if run in Ray Cluster, this method is run remote with gpu resources requested if available
def memory_tune_config(config, dataset, model_category, row_count, backend):
    backend = initialize_backend(backend)

    fits_in_memory = False
    tried_reduce_seq_len = False
    config_obj = ModelConfig.from_dict(config)
    raw_config = config_obj.to_dict()
    training_set_metadata = get_trainingset_metadata(raw_config, dataset, backend)
    modified_hyperparam_search_space = copy.deepcopy(raw_config[HYPEROPT]["parameters"])
    current_param_values = {}
    param_list = []
    model_type = get_model_type(raw_config)
    if model_type in RANKED_MODIFIABLE_PARAM_LIST:
        params_to_modify = RANKED_MODIFIABLE_PARAM_LIST[model_type]
        if len(params_to_modify.keys()) > 0:
            param_list = list(params_to_modify.keys())
            max_memory = _get_machine_memory()
            initialize_pytorch()

    while param_list:
        # compute memory utilization
        current_param_values = get_new_params(current_param_values, modified_hyperparam_search_space, params_to_modify)
        temp_config = sub_new_params(raw_config, current_param_values)
        config_obj = ModelConfig.from_dict(temp_config)
        mem_use = compute_memory_usage(config_obj, training_set_metadata, model_category)
        if mem_use > max_memory and model_category == TEXT and not tried_reduce_seq_len:
            tried_reduce_seq_len = True
            if reduce_text_feature_max_length(config, training_set_metadata):
                reduce_text_feature_max_length(temp_config, training_set_metadata)
                config_obj = ModelConfig.from_dict(temp_config)
                mem_use = compute_memory_usage(config_obj, training_set_metadata, model_category)
        logger.info(f"Checking model estimated mem use {mem_use} against memory size {max_memory}")
        if mem_use <= max_memory:
            fits_in_memory = True
            break
        # check if we have exhausted tuning of current param (e.g. we can no longer reduce the param value)
        param, min_value = param_list[0], params_to_modify[param_list[0]]

        if param in modified_hyperparam_search_space.keys():
            param_space = modified_hyperparam_search_space[param]["space"]
            if param_space == "choice":
                if (
                    len(modified_hyperparam_search_space[param]["categories"]) >= 2
                    and modified_hyperparam_search_space[param]["categories"][-2] >= min_value
                ):
                    modified_hyperparam_search_space[param]["categories"] = modified_hyperparam_search_space[param][
                        "categories"
                    ][:-1]
                else:
                    param_list.pop(0)  # exhausted reduction of this parameter
            else:
                # reduce by 10%
                upper_bound, lower_bound = (
                    modified_hyperparam_search_space[param]["upper"],
                    modified_hyperparam_search_space[param]["lower"],
                )
                reduction_val = (upper_bound - lower_bound) * 0.1
                new_upper_bound = upper_bound - reduction_val
                if (new_upper_bound) > lower_bound and new_upper_bound > min_value:
                    modified_hyperparam_search_space[param]["upper"] = new_upper_bound
                else:
                    param_list.pop(0)  # exhausted reduction of this parameter
        else:
            param_list.pop(0)  # param not in hyperopt search space

    if model_category == TEXT and row_count > AUTOML_LARGE_TEXT_DATASET:
        if "checkpoints_per_epoch" not in config[TRAINER] and "steps_per_checkpoint" not in config[TRAINER]:
            checkpoints_per_epoch = max(2, math.floor(row_count / AUTOML_MAX_ROWS_PER_CHECKPOINT))
            config[TRAINER][
                "checkpoints_per_epoch"
            ] = checkpoints_per_epoch  # decrease latency to get model accuracy signal
        if "evaluate_training_set" not in config[TRAINER]:
            config[TRAINER]["evaluate_training_set"] = False  # reduce overhead for increased evaluation frequency
        if not fits_in_memory:
            # Switch to smaller pre-trained model encoder for large datasets.
            _update_text_encoder(config["input_features"], AUTOML_DEFAULT_TEXT_ENCODER, AUTOML_SMALLER_TEXT_ENCODER)

    modified_config = copy.deepcopy(config)

    modified_config[HYPEROPT]["parameters"] = modified_hyperparam_search_space
    modified_config[HYPEROPT]["executor"]["num_samples"] = _update_num_samples(
        modified_config[HYPEROPT]["executor"]["num_samples"], modified_hyperparam_search_space
    )
    return modified_config, fits_in_memory



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ludwig/profiling/proto/whylogs_messages.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2
from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='ludwig/profiling/proto/whylogs_messages.proto',
  package='dataset_profile',
  syntax='proto3',
  serialized_options=b'Z+github.com/ludwig-ai/ludwig/dataset_profile',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n-ludwig/profiling/proto/whylogs_messages.proto\x12\x0f\x64\x61taset_profile\x1a\x19google/protobuf/any.proto\x1a\x1cgoogle/protobuf/struct.proto\"\x8e\x01\n\x08\x44\x61taType\x12,\n\x04type\x18\x01 \x01(\x0e\x32\x1e.dataset_profile.DataType.Type\"T\n\x04Type\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x08\n\x04NULL\x10\x01\x12\x0e\n\nFRACTIONAL\x10\x02\x12\x0c\n\x08INTEGRAL\x10\x03\x12\x0b\n\x07\x42OOLEAN\x10\x04\x12\n\n\x06STRING\x10\x05\"\"\n\x10HllSketchMessage\x12\x0e\n\x06sketch\x18\x01 \x01(\x0c\",\n\x1a\x46requentItemsSketchMessage\x12\x0e\n\x06sketch\x18\x01 \x01(\x0c\"\"\n\x10KllSketchMessage\x12\x0e\n\x06sketch\x18\x01 \x01(\x0c\"\"\n\x10\x43pcSketchMessage\x12\x0e\n\x06sketch\x18\x01 \x01(\x0c\"\x9e\x03\n\x16MetricComponentMessage\x12\x0f\n\x07type_id\x18\x01 \x01(\r\x12\x0b\n\x01n\x18\x02 \x01(\x03H\x00\x12\x0b\n\x01\x64\x18\x03 \x01(\x01H\x00\x12\x45\n\x0e\x66requent_items\x18\x04 \x01(\x0b\x32+.dataset_profile.FrequentItemsSketchMessageH\x00\x12\x30\n\x03hll\x18\x05 \x01(\x0b\x32!.dataset_profile.HllSketchMessageH\x00\x12\x30\n\x03kll\x18\x06 \x01(\x0b\x32!.dataset_profile.KllSketchMessageH\x00\x12\x30\n\x03\x63pc\x18\x07 \x01(\x0b\x32!.dataset_profile.CpcSketchMessageH\x00\x12\x1a\n\x10serialized_bytes\x18\n \x01(\x0cH\x00\x12\x32\n\x0f\x64\x61taclass_param\x18\x0b \x01(\x0b\x32\x17.google.protobuf.StructH\x00\x12#\n\x03msg\x18\x0c \x01(\x0b\x32\x14.google.protobuf.AnyH\x00\x42\x07\n\x05value\"\xc2\x01\n\rMetricMessage\x12O\n\x11metric_components\x18\x01 \x03(\x0b\x32\x34.dataset_profile.MetricMessage.MetricComponentsEntry\x1a`\n\x15MetricComponentsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x36\n\x05value\x18\x02 \x01(\x0b\x32\'.dataset_profile.MetricComponentMessage:\x02\x38\x01\"\xc2\x01\n\rColumnMessage\x12O\n\x11metric_components\x18\x01 \x03(\x0b\x32\x34.dataset_profile.ColumnMessage.MetricComponentsEntry\x1a`\n\x15MetricComponentsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x36\n\x05value\x18\x02 \x01(\x0b\x32\'.dataset_profile.MetricComponentMessage:\x02\x38\x01\"\xe4\x02\n\x11\x44\x61tasetProperties\x12\x1c\n\x14schema_major_version\x18\x01 \x01(\r\x12\x1c\n\x14schema_minor_version\x18\x02 \x01(\r\x12\x1a\n\x12\x63reation_timestamp\x18\x04 \x01(\x04\x12\x19\n\x11\x64\x61taset_timestamp\x18\x05 \x01(\x04\x12:\n\x04tags\x18\x06 \x03(\x0b\x32,.dataset_profile.DatasetProperties.TagsEntry\x12\x42\n\x08metadata\x18\x07 \x03(\x0b\x32\x30.dataset_profile.DatasetProperties.MetadataEntry\x1a+\n\tTagsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a/\n\rMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\x1f\n\x0c\x43hunkOffsets\x12\x0f\n\x07offsets\x18\x01 \x03(\x04\"\xc0\x01\n\x0c\x43hunkMessage\x12N\n\x11metric_components\x18\x01 \x03(\x0b\x32\x33.dataset_profile.ChunkMessage.MetricComponentsEntry\x1a`\n\x15MetricComponentsEntry\x12\x0b\n\x03key\x18\x01 \x01(\r\x12\x36\n\x05value\x18\x02 \x01(\x0b\x32\'.dataset_profile.MetricComponentMessage:\x02\x38\x01\"\x86\x01\n\x0b\x43hunkHeader\x12\x34\n\x04type\x18\x01 \x01(\x0e\x32&.dataset_profile.ChunkHeader.ChunkType\x12\x0b\n\x03tag\x18\x02 \x01(\t\x12\x0e\n\x06length\x18\x03 \x01(\r\"$\n\tChunkType\x12\x0b\n\x07\x44\x41TASET\x10\x00\x12\n\n\x06\x43OLUMN\x10\x02\"\xd4\x03\n\x14\x44\x61tasetProfileHeader\x12\x36\n\nproperties\x18\x01 \x01(\x0b\x32\".dataset_profile.DatasetProperties\x12P\n\x0e\x63olumn_offsets\x18\x02 \x03(\x0b\x32\x38.dataset_profile.DatasetProfileHeader.ColumnOffsetsEntry\x12\x35\n\x0emetric_offsets\x18\x03 \x03(\x0b\x32\x1d.dataset_profile.ChunkOffsets\x12\x0e\n\x06length\x18\x04 \x01(\x04\x12[\n\x14indexed_metric_paths\x18\x05 \x03(\x0b\x32=.dataset_profile.DatasetProfileHeader.IndexedMetricPathsEntry\x1aS\n\x12\x43olumnOffsetsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12,\n\x05value\x18\x02 \x01(\x0b\x32\x1d.dataset_profile.ChunkOffsets:\x02\x38\x01\x1a\x39\n\x17IndexedMetricPathsEntry\x12\x0b\n\x03key\x18\x01 \x01(\r\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"(\n\nSegmentTag\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"4\n\x07Segment\x12)\n\x04tags\x18\x01 \x03(\x0b\x32\x1b.dataset_profile.SegmentTag\"\xcd\x01\n\x14\x44\x61tasetSegmentHeader\x12\x14\n\x0chas_segments\x18\x01 \x01(\x08\x12*\n\x08segments\x18\x04 \x03(\x0b\x32\x18.dataset_profile.Segment\x12\x43\n\x07offsets\x18\x05 \x03(\x0b\x32\x32.dataset_profile.DatasetSegmentHeader.OffsetsEntry\x1a.\n\x0cOffsetsEntry\x12\x0b\n\x03key\x18\x01 \x01(\r\x12\r\n\x05value\x18\x02 \x01(\x04:\x02\x38\x01\x42-Z+github.com/ludwig-ai/ludwig/dataset_profileb\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_any__pb2.DESCRIPTOR,google_dot_protobuf_dot_struct__pb2.DESCRIPTOR,])



_DATATYPE_TYPE = _descriptor.EnumDescriptor(
  name='Type',
  full_name='dataset_profile.DataType.Type',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='NULL', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FRACTIONAL', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INTEGRAL', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='BOOLEAN', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='STRING', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=182,
  serialized_end=266,
)
_sym_db.RegisterEnumDescriptor(_DATATYPE_TYPE)

_CHUNKHEADER_CHUNKTYPE = _descriptor.EnumDescriptor(
  name='ChunkType',
  full_name='dataset_profile.ChunkHeader.ChunkType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='DATASET', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='COLUMN', index=1, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1919,
  serialized_end=1955,
)
_sym_db.RegisterEnumDescriptor(_CHUNKHEADER_CHUNKTYPE)


_DATATYPE = _descriptor.Descriptor(
  name='DataType',
  full_name='dataset_profile.DataType',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='dataset_profile.DataType.type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _DATATYPE_TYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=124,
  serialized_end=266,
)


_HLLSKETCHMESSAGE = _descriptor.Descriptor(
  name='HllSketchMessage',
  full_name='dataset_profile.HllSketchMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='sketch', full_name='dataset_profile.HllSketchMessage.sketch', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=268,
  serialized_end=302,
)


_FREQUENTITEMSSKETCHMESSAGE = _descriptor.Descriptor(
  name='FrequentItemsSketchMessage',
  full_name='dataset_profile.FrequentItemsSketchMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='sketch', full_name='dataset_profile.FrequentItemsSketchMessage.sketch', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=304,
  serialized_end=348,
)


_KLLSKETCHMESSAGE = _descriptor.Descriptor(
  name='KllSketchMessage',
  full_name='dataset_profile.KllSketchMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='sketch', full_name='dataset_profile.KllSketchMessage.sketch', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=350,
  serialized_end=384,
)


_CPCSKETCHMESSAGE = _descriptor.Descriptor(
  name='CpcSketchMessage',
  full_name='dataset_profile.CpcSketchMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='sketch', full_name='dataset_profile.CpcSketchMessage.sketch', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=386,
  serialized_end=420,
)


_METRICCOMPONENTMESSAGE = _descriptor.Descriptor(
  name='MetricComponentMessage',
  full_name='dataset_profile.MetricComponentMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='type_id', full_name='dataset_profile.MetricComponentMessage.type_id', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='n', full_name='dataset_profile.MetricComponentMessage.n', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='d', full_name='dataset_profile.MetricComponentMessage.d', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='frequent_items', full_name='dataset_profile.MetricComponentMessage.frequent_items', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='hll', full_name='dataset_profile.MetricComponentMessage.hll', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='kll', full_name='dataset_profile.MetricComponentMessage.kll', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='cpc', full_name='dataset_profile.MetricComponentMessage.cpc', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='serialized_bytes', full_name='dataset_profile.MetricComponentMessage.serialized_bytes', index=7,
      number=10, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dataclass_param', full_name='dataset_profile.MetricComponentMessage.dataclass_param', index=8,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='msg', full_name='dataset_profile.MetricComponentMessage.msg', index=9,
      number=12, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='value', full_name='dataset_profile.MetricComponentMessage.value',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=423,
  serialized_end=837,
)


_METRICMESSAGE_METRICCOMPONENTSENTRY = _descriptor.Descriptor(
  name='MetricComponentsEntry',
  full_name='dataset_profile.MetricMessage.MetricComponentsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='dataset_profile.MetricMessage.MetricComponentsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='dataset_profile.MetricMessage.MetricComponentsEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=938,
  serialized_end=1034,
)

_METRICMESSAGE = _descriptor.Descriptor(
  name='MetricMessage',
  full_name='dataset_profile.MetricMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='metric_components', full_name='dataset_profile.MetricMessage.metric_components', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_METRICMESSAGE_METRICCOMPONENTSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=840,
  serialized_end=1034,
)


_COLUMNMESSAGE_METRICCOMPONENTSENTRY = _descriptor.Descriptor(
  name='MetricComponentsEntry',
  full_name='dataset_profile.ColumnMessage.MetricComponentsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='dataset_profile.ColumnMessage.MetricComponentsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='dataset_profile.ColumnMessage.MetricComponentsEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=938,
  serialized_end=1034,
)

_COLUMNMESSAGE = _descriptor.Descriptor(
  name='ColumnMessage',
  full_name='dataset_profile.ColumnMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='metric_components', full_name='dataset_profile.ColumnMessage.metric_components', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_COLUMNMESSAGE_METRICCOMPONENTSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1037,
  serialized_end=1231,
)


_DATASETPROPERTIES_TAGSENTRY = _descriptor.Descriptor(
  name='TagsEntry',
  full_name='dataset_profile.DatasetProperties.TagsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='dataset_profile.DatasetProperties.TagsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='dataset_profile.DatasetProperties.TagsEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1498,
  serialized_end=1541,
)

_DATASETPROPERTIES_METADATAENTRY = _descriptor.Descriptor(
  name='MetadataEntry',
  full_name='dataset_profile.DatasetProperties.MetadataEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='dataset_profile.DatasetProperties.MetadataEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='dataset_profile.DatasetProperties.MetadataEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1543,
  serialized_end=1590,
)

_DATASETPROPERTIES = _descriptor.Descriptor(
  name='DatasetProperties',
  full_name='dataset_profile.DatasetProperties',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='schema_major_version', full_name='dataset_profile.DatasetProperties.schema_major_version', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='schema_minor_version', full_name='dataset_profile.DatasetProperties.schema_minor_version', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='creation_timestamp', full_name='dataset_profile.DatasetProperties.creation_timestamp', index=2,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dataset_timestamp', full_name='dataset_profile.DatasetProperties.dataset_timestamp', index=3,
      number=5, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='tags', full_name='dataset_profile.DatasetProperties.tags', index=4,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='metadata', full_name='dataset_profile.DatasetProperties.metadata', index=5,
      number=7, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_DATASETPROPERTIES_TAGSENTRY, _DATASETPROPERTIES_METADATAENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1234,
  serialized_end=1590,
)


_CHUNKOFFSETS = _descriptor.Descriptor(
  name='ChunkOffsets',
  full_name='dataset_profile.ChunkOffsets',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='offsets', full_name='dataset_profile.ChunkOffsets.offsets', index=0,
      number=1, type=4, cpp_type=4, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1592,
  serialized_end=1623,
)


_CHUNKMESSAGE_METRICCOMPONENTSENTRY = _descriptor.Descriptor(
  name='MetricComponentsEntry',
  full_name='dataset_profile.ChunkMessage.MetricComponentsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='dataset_profile.ChunkMessage.MetricComponentsEntry.key', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='dataset_profile.ChunkMessage.MetricComponentsEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1722,
  serialized_end=1818,
)

_CHUNKMESSAGE = _descriptor.Descriptor(
  name='ChunkMessage',
  full_name='dataset_profile.ChunkMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='metric_components', full_name='dataset_profile.ChunkMessage.metric_components', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_CHUNKMESSAGE_METRICCOMPONENTSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1626,
  serialized_end=1818,
)


_CHUNKHEADER = _descriptor.Descriptor(
  name='ChunkHeader',
  full_name='dataset_profile.ChunkHeader',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='dataset_profile.ChunkHeader.type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='tag', full_name='dataset_profile.ChunkHeader.tag', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='length', full_name='dataset_profile.ChunkHeader.length', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _CHUNKHEADER_CHUNKTYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1821,
  serialized_end=1955,
)


_DATASETPROFILEHEADER_COLUMNOFFSETSENTRY = _descriptor.Descriptor(
  name='ColumnOffsetsEntry',
  full_name='dataset_profile.DatasetProfileHeader.ColumnOffsetsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='dataset_profile.DatasetProfileHeader.ColumnOffsetsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='dataset_profile.DatasetProfileHeader.ColumnOffsetsEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2284,
  serialized_end=2367,
)

_DATASETPROFILEHEADER_INDEXEDMETRICPATHSENTRY = _descriptor.Descriptor(
  name='IndexedMetricPathsEntry',
  full_name='dataset_profile.DatasetProfileHeader.IndexedMetricPathsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='dataset_profile.DatasetProfileHeader.IndexedMetricPathsEntry.key', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='dataset_profile.DatasetProfileHeader.IndexedMetricPathsEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2369,
  serialized_end=2426,
)

_DATASETPROFILEHEADER = _descriptor.Descriptor(
  name='DatasetProfileHeader',
  full_name='dataset_profile.DatasetProfileHeader',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='properties', full_name='dataset_profile.DatasetProfileHeader.properties', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='column_offsets', full_name='dataset_profile.DatasetProfileHeader.column_offsets', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='metric_offsets', full_name='dataset_profile.DatasetProfileHeader.metric_offsets', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='length', full_name='dataset_profile.DatasetProfileHeader.length', index=3,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='indexed_metric_paths', full_name='dataset_profile.DatasetProfileHeader.indexed_metric_paths', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_DATASETPROFILEHEADER_COLUMNOFFSETSENTRY, _DATASETPROFILEHEADER_INDEXEDMETRICPATHSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1958,
  serialized_end=2426,
)


_SEGMENTTAG = _descriptor.Descriptor(
  name='SegmentTag',
  full_name='dataset_profile.SegmentTag',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='dataset_profile.SegmentTag.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='dataset_profile.SegmentTag.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2428,
  serialized_end=2468,
)


_SEGMENT = _descriptor.Descriptor(
  name='Segment',
  full_name='dataset_profile.Segment',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='tags', full_name='dataset_profile.Segment.tags', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2470,
  serialized_end=2522,
)


_DATASETSEGMENTHEADER_OFFSETSENTRY = _descriptor.Descriptor(
  name='OffsetsEntry',
  full_name='dataset_profile.DatasetSegmentHeader.OffsetsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='dataset_profile.DatasetSegmentHeader.OffsetsEntry.key', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='dataset_profile.DatasetSegmentHeader.OffsetsEntry.value', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2684,
  serialized_end=2730,
)

_DATASETSEGMENTHEADER = _descriptor.Descriptor(
  name='DatasetSegmentHeader',
  full_name='dataset_profile.DatasetSegmentHeader',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='has_segments', full_name='dataset_profile.DatasetSegmentHeader.has_segments', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='segments', full_name='dataset_profile.DatasetSegmentHeader.segments', index=1,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='offsets', full_name='dataset_profile.DatasetSegmentHeader.offsets', index=2,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_DATASETSEGMENTHEADER_OFFSETSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2525,
  serialized_end=2730,
)

_DATATYPE.fields_by_name['type'].enum_type = _DATATYPE_TYPE
_DATATYPE_TYPE.containing_type = _DATATYPE
_METRICCOMPONENTMESSAGE.fields_by_name['frequent_items'].message_type = _FREQUENTITEMSSKETCHMESSAGE
_METRICCOMPONENTMESSAGE.fields_by_name['hll'].message_type = _HLLSKETCHMESSAGE
_METRICCOMPONENTMESSAGE.fields_by_name['kll'].message_type = _KLLSKETCHMESSAGE
_METRICCOMPONENTMESSAGE.fields_by_name['cpc'].message_type = _CPCSKETCHMESSAGE
_METRICCOMPONENTMESSAGE.fields_by_name['dataclass_param'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_METRICCOMPONENTMESSAGE.fields_by_name['msg'].message_type = google_dot_protobuf_dot_any__pb2._ANY
_METRICCOMPONENTMESSAGE.oneofs_by_name['value'].fields.append(
  _METRICCOMPONENTMESSAGE.fields_by_name['n'])
_METRICCOMPONENTMESSAGE.fields_by_name['n'].containing_oneof = _METRICCOMPONENTMESSAGE.oneofs_by_name['value']
_METRICCOMPONENTMESSAGE.oneofs_by_name['value'].fields.append(
  _METRICCOMPONENTMESSAGE.fields_by_name['d'])
_METRICCOMPONENTMESSAGE.fields_by_name['d'].containing_oneof = _METRICCOMPONENTMESSAGE.oneofs_by_name['value']
_METRICCOMPONENTMESSAGE.oneofs_by_name['value'].fields.append(
  _METRICCOMPONENTMESSAGE.fields_by_name['frequent_items'])
_METRICCOMPONENTMESSAGE.fields_by_name['frequent_items'].containing_oneof = _METRICCOMPONENTMESSAGE.oneofs_by_name['value']
_METRICCOMPONENTMESSAGE.oneofs_by_name['value'].fields.append(
  _METRICCOMPONENTMESSAGE.fields_by_name['hll'])
_METRICCOMPONENTMESSAGE.fields_by_name['hll'].containing_oneof = _METRICCOMPONENTMESSAGE.oneofs_by_name['value']
_METRICCOMPONENTMESSAGE.oneofs_by_name['value'].fields.append(
  _METRICCOMPONENTMESSAGE.fields_by_name['kll'])
_METRICCOMPONENTMESSAGE.fields_by_name['kll'].containing_oneof = _METRICCOMPONENTMESSAGE.oneofs_by_name['value']
_METRICCOMPONENTMESSAGE.oneofs_by_name['value'].fields.append(
  _METRICCOMPONENTMESSAGE.fields_by_name['cpc'])
_METRICCOMPONENTMESSAGE.fields_by_name['cpc'].containing_oneof = _METRICCOMPONENTMESSAGE.oneofs_by_name['value']
_METRICCOMPONENTMESSAGE.oneofs_by_name['value'].fields.append(
  _METRICCOMPONENTMESSAGE.fields_by_name['serialized_bytes'])
_METRICCOMPONENTMESSAGE.fields_by_name['serialized_bytes'].containing_oneof = _METRICCOMPONENTMESSAGE.oneofs_by_name['value']
_METRICCOMPONENTMESSAGE.oneofs_by_name['value'].fields.append(
  _METRICCOMPONENTMESSAGE.fields_by_name['dataclass_param'])
_METRICCOMPONENTMESSAGE.fields_by_name['dataclass_param'].containing_oneof = _METRICCOMPONENTMESSAGE.oneofs_by_name['value']
_METRICCOMPONENTMESSAGE.oneofs_by_name['value'].fields.append(
  _METRICCOMPONENTMESSAGE.fields_by_name['msg'])
_METRICCOMPONENTMESSAGE.fields_by_name['msg'].containing_oneof = _METRICCOMPONENTMESSAGE.oneofs_by_name['value']
_METRICMESSAGE_METRICCOMPONENTSENTRY.fields_by_name['value'].message_type = _METRICCOMPONENTMESSAGE
_METRICMESSAGE_METRICCOMPONENTSENTRY.containing_type = _METRICMESSAGE
_METRICMESSAGE.fields_by_name['metric_components'].message_type = _METRICMESSAGE_METRICCOMPONENTSENTRY
_COLUMNMESSAGE_METRICCOMPONENTSENTRY.fields_by_name['value'].message_type = _METRICCOMPONENTMESSAGE
_COLUMNMESSAGE_METRICCOMPONENTSENTRY.containing_type = _COLUMNMESSAGE
_COLUMNMESSAGE.fields_by_name['metric_components'].message_type = _COLUMNMESSAGE_METRICCOMPONENTSENTRY
_DATASETPROPERTIES_TAGSENTRY.containing_type = _DATASETPROPERTIES
_DATASETPROPERTIES_METADATAENTRY.containing_type = _DATASETPROPERTIES
_DATASETPROPERTIES.fields_by_name['tags'].message_type = _DATASETPROPERTIES_TAGSENTRY
_DATASETPROPERTIES.fields_by_name['metadata'].message_type = _DATASETPROPERTIES_METADATAENTRY
_CHUNKMESSAGE_METRICCOMPONENTSENTRY.fields_by_name['value'].message_type = _METRICCOMPONENTMESSAGE
_CHUNKMESSAGE_METRICCOMPONENTSENTRY.containing_type = _CHUNKMESSAGE
_CHUNKMESSAGE.fields_by_name['metric_components'].message_type = _CHUNKMESSAGE_METRICCOMPONENTSENTRY
_CHUNKHEADER.fields_by_name['type'].enum_type = _CHUNKHEADER_CHUNKTYPE
_CHUNKHEADER_CHUNKTYPE.containing_type = _CHUNKHEADER
_DATASETPROFILEHEADER_COLUMNOFFSETSENTRY.fields_by_name['value'].message_type = _CHUNKOFFSETS
_DATASETPROFILEHEADER_COLUMNOFFSETSENTRY.containing_type = _DATASETPROFILEHEADER
_DATASETPROFILEHEADER_INDEXEDMETRICPATHSENTRY.containing_type = _DATASETPROFILEHEADER
_DATASETPROFILEHEADER.fields_by_name['properties'].message_type = _DATASETPROPERTIES
_DATASETPROFILEHEADER.fields_by_name['column_offsets'].message_type = _DATASETPROFILEHEADER_COLUMNOFFSETSENTRY
_DATASETPROFILEHEADER.fields_by_name['metric_offsets'].message_type = _CHUNKOFFSETS
_DATASETPROFILEHEADER.fields_by_name['indexed_metric_paths'].message_type = _DATASETPROFILEHEADER_INDEXEDMETRICPATHSENTRY
_SEGMENT.fields_by_name['tags'].message_type = _SEGMENTTAG
_DATASETSEGMENTHEADER_OFFSETSENTRY.containing_type = _DATASETSEGMENTHEADER
_DATASETSEGMENTHEADER.fields_by_name['segments'].message_type = _SEGMENT
_DATASETSEGMENTHEADER.fields_by_name['offsets'].message_type = _DATASETSEGMENTHEADER_OFFSETSENTRY
DESCRIPTOR.message_types_by_name['DataType'] = _DATATYPE
DESCRIPTOR.message_types_by_name['HllSketchMessage'] = _HLLSKETCHMESSAGE
DESCRIPTOR.message_types_by_name['FrequentItemsSketchMessage'] = _FREQUENTITEMSSKETCHMESSAGE
DESCRIPTOR.message_types_by_name['KllSketchMessage'] = _KLLSKETCHMESSAGE
DESCRIPTOR.message_types_by_name['CpcSketchMessage'] = _CPCSKETCHMESSAGE
DESCRIPTOR.message_types_by_name['MetricComponentMessage'] = _METRICCOMPONENTMESSAGE
DESCRIPTOR.message_types_by_name['MetricMessage'] = _METRICMESSAGE
DESCRIPTOR.message_types_by_name['ColumnMessage'] = _COLUMNMESSAGE
DESCRIPTOR.message_types_by_name['DatasetProperties'] = _DATASETPROPERTIES
DESCRIPTOR.message_types_by_name['ChunkOffsets'] = _CHUNKOFFSETS
DESCRIPTOR.message_types_by_name['ChunkMessage'] = _CHUNKMESSAGE
DESCRIPTOR.message_types_by_name['ChunkHeader'] = _CHUNKHEADER
DESCRIPTOR.message_types_by_name['DatasetProfileHeader'] = _DATASETPROFILEHEADER
DESCRIPTOR.message_types_by_name['SegmentTag'] = _SEGMENTTAG
DESCRIPTOR.message_types_by_name['Segment'] = _SEGMENT
DESCRIPTOR.message_types_by_name['DatasetSegmentHeader'] = _DATASETSEGMENTHEADER
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

DataType = _reflection.GeneratedProtocolMessageType('DataType', (_message.Message,), {
  'DESCRIPTOR' : _DATATYPE,
  '__module__' : 'ludwig.profiling.proto.whylogs_messages_pb2'
  # @@protoc_insertion_point(class_scope:dataset_profile.DataType)
  })
_sym_db.RegisterMessage(DataType)

HllSketchMessage = _reflection.GeneratedProtocolMessageType('HllSketchMessage', (_message.Message,), {
  'DESCRIPTOR' : _HLLSKETCHMESSAGE,
  '__module__' : 'ludwig.profiling.proto.whylogs_messages_pb2'
  # @@protoc_insertion_point(class_scope:dataset_profile.HllSketchMessage)
  })
_sym_db.RegisterMessage(HllSketchMessage)

FrequentItemsSketchMessage = _reflection.GeneratedProtocolMessageType('FrequentItemsSketchMessage', (_message.Message,), {
  'DESCRIPTOR' : _FREQUENTITEMSSKETCHMESSAGE,
  '__module__' : 'ludwig.profiling.proto.whylogs_messages_pb2'
  # @@protoc_insertion_point(class_scope:dataset_profile.FrequentItemsSketchMessage)
  })
_sym_db.RegisterMessage(FrequentItemsSketchMessage)

KllSketchMessage = _reflection.GeneratedProtocolMessageType('KllSketchMessage', (_message.Message,), {
  'DESCRIPTOR' : _KLLSKETCHMESSAGE,
  '__module__' : 'ludwig.profiling.proto.whylogs_messages_pb2'
  # @@protoc_insertion_point(class_scope:dataset_profile.KllSketchMessage)
  })
_sym_db.RegisterMessage(KllSketchMessage)

CpcSketchMessage = _reflection.GeneratedProtocolMessageType('CpcSketchMessage', (_message.Message,), {
  'DESCRIPTOR' : _CPCSKETCHMESSAGE,
  '__module__' : 'ludwig.profiling.proto.whylogs_messages_pb2'
  # @@protoc_insertion_point(class_scope:dataset_profile.CpcSketchMessage)
  })
_sym_db.RegisterMessage(CpcSketchMessage)

MetricComponentMessage = _reflection.GeneratedProtocolMessageType('MetricComponentMessage', (_message.Message,), {
  'DESCRIPTOR' : _METRICCOMPONENTMESSAGE,
  '__module__' : 'ludwig.profiling.proto.whylogs_messages_pb2'
  # @@protoc_insertion_point(class_scope:dataset_profile.MetricComponentMessage)
  })
_sym_db.RegisterMessage(MetricComponentMessage)

MetricMessage = _reflection.GeneratedProtocolMessageType('MetricMessage', (_message.Message,), {

  'MetricComponentsEntry' : _reflection.GeneratedProtocolMessageType('MetricComponentsEntry', (_message.Message,), {
    'DESCRIPTOR' : _METRICMESSAGE_METRICCOMPONENTSENTRY,
    '__module__' : 'ludwig.profiling.proto.whylogs_messages_pb2'
    # @@protoc_insertion_point(class_scope:dataset_profile.MetricMessage.MetricComponentsEntry)
    })
  ,
  'DESCRIPTOR' : _METRICMESSAGE,
  '__module__' : 'ludwig.profiling.proto.whylogs_messages_pb2'
  # @@protoc_insertion_point(class_scope:dataset_profile.MetricMessage)
  })
_sym_db.RegisterMessage(MetricMessage)
_sym_db.RegisterMessage(MetricMessage.MetricComponentsEntry)

ColumnMessage = _reflection.GeneratedProtocolMessageType('ColumnMessage', (_message.Message,), {

  'MetricComponentsEntry' : _reflection.GeneratedProtocolMessageType('MetricComponentsEntry', (_message.Message,), {
    'DESCRIPTOR' : _COLUMNMESSAGE_METRICCOMPONENTSENTRY,
    '__module__' : 'ludwig.profiling.proto.whylogs_messages_pb2'
    # @@protoc_insertion_point(class_scope:dataset_profile.ColumnMessage.MetricComponentsEntry)
    })
  ,
  'DESCRIPTOR' : _COLUMNMESSAGE,
  '__module__' : 'ludwig.profiling.proto.whylogs_messages_pb2'
  # @@protoc_insertion_point(class_scope:dataset_profile.ColumnMessage)
  })
_sym_db.RegisterMessage(ColumnMessage)
_sym_db.RegisterMessage(ColumnMessage.MetricComponentsEntry)

DatasetProperties = _reflection.GeneratedProtocolMessageType('DatasetProperties', (_message.Message,), {

  'TagsEntry' : _reflection.GeneratedProtocolMessageType('TagsEntry', (_message.Message,), {
    'DESCRIPTOR' : _DATASETPROPERTIES_TAGSENTRY,
    '__module__' : 'ludwig.profiling.proto.whylogs_messages_pb2'
    # @@protoc_insertion_point(class_scope:dataset_profile.DatasetProperties.TagsEntry)
    })
  ,

  'MetadataEntry' : _reflection.GeneratedProtocolMessageType('MetadataEntry', (_message.Message,), {
    'DESCRIPTOR' : _DATASETPROPERTIES_METADATAENTRY,
    '__module__' : 'ludwig.profiling.proto.whylogs_messages_pb2'
    # @@protoc_insertion_point(class_scope:dataset_profile.DatasetProperties.MetadataEntry)
    })
  ,
  'DESCRIPTOR' : _DATASETPROPERTIES,
  '__module__' : 'ludwig.profiling.proto.whylogs_messages_pb2'
  # @@protoc_insertion_point(class_scope:dataset_profile.DatasetProperties)
  })
_sym_db.RegisterMessage(DatasetProperties)
_sym_db.RegisterMessage(DatasetProperties.TagsEntry)
_sym_db.RegisterMessage(DatasetProperties.MetadataEntry)

ChunkOffsets = _reflection.GeneratedProtocolMessageType('ChunkOffsets', (_message.Message,), {
  'DESCRIPTOR' : _CHUNKOFFSETS,
  '__module__' : 'ludwig.profiling.proto.whylogs_messages_pb2'
  # @@protoc_insertion_point(class_scope:dataset_profile.ChunkOffsets)
  })
_sym_db.RegisterMessage(ChunkOffsets)

ChunkMessage = _reflection.GeneratedProtocolMessageType('ChunkMessage', (_message.Message,), {

  'MetricComponentsEntry' : _reflection.GeneratedProtocolMessageType('MetricComponentsEntry', (_message.Message,), {
    'DESCRIPTOR' : _CHUNKMESSAGE_METRICCOMPONENTSENTRY,
    '__module__' : 'ludwig.profiling.proto.whylogs_messages_pb2'
    # @@protoc_insertion_point(class_scope:dataset_profile.ChunkMessage.MetricComponentsEntry)
    })
  ,
  'DESCRIPTOR' : _CHUNKMESSAGE,
  '__module__' : 'ludwig.profiling.proto.whylogs_messages_pb2'
  # @@protoc_insertion_point(class_scope:dataset_profile.ChunkMessage)
  })
_sym_db.RegisterMessage(ChunkMessage)
_sym_db.RegisterMessage(ChunkMessage.MetricComponentsEntry)

ChunkHeader = _reflection.GeneratedProtocolMessageType('ChunkHeader', (_message.Message,), {
  'DESCRIPTOR' : _CHUNKHEADER,
  '__module__' : 'ludwig.profiling.proto.whylogs_messages_pb2'
  # @@protoc_insertion_point(class_scope:dataset_profile.ChunkHeader)
  })
_sym_db.RegisterMessage(ChunkHeader)

DatasetProfileHeader = _reflection.GeneratedProtocolMessageType('DatasetProfileHeader', (_message.Message,), {

  'ColumnOffsetsEntry' : _reflection.GeneratedProtocolMessageType('ColumnOffsetsEntry', (_message.Message,), {
    'DESCRIPTOR' : _DATASETPROFILEHEADER_COLUMNOFFSETSENTRY,
    '__module__' : 'ludwig.profiling.proto.whylogs_messages_pb2'
    # @@protoc_insertion_point(class_scope:dataset_profile.DatasetProfileHeader.ColumnOffsetsEntry)
    })
  ,

  'IndexedMetricPathsEntry' : _reflection.GeneratedProtocolMessageType('IndexedMetricPathsEntry', (_message.Message,), {
    'DESCRIPTOR' : _DATASETPROFILEHEADER_INDEXEDMETRICPATHSENTRY,
    '__module__' : 'ludwig.profiling.proto.whylogs_messages_pb2'
    # @@protoc_insertion_point(class_scope:dataset_profile.DatasetProfileHeader.IndexedMetricPathsEntry)
    })
  ,
  'DESCRIPTOR' : _DATASETPROFILEHEADER,
  '__module__' : 'ludwig.profiling.proto.whylogs_messages_pb2'
  # @@protoc_insertion_point(class_scope:dataset_profile.DatasetProfileHeader)
  })
_sym_db.RegisterMessage(DatasetProfileHeader)
_sym_db.RegisterMessage(DatasetProfileHeader.ColumnOffsetsEntry)
_sym_db.RegisterMessage(DatasetProfileHeader.IndexedMetricPathsEntry)

SegmentTag = _reflection.GeneratedProtocolMessageType('SegmentTag', (_message.Message,), {
  'DESCRIPTOR' : _SEGMENTTAG,
  '__module__' : 'ludwig.profiling.proto.whylogs_messages_pb2'
  # @@protoc_insertion_point(class_scope:dataset_profile.SegmentTag)
  })
_sym_db.RegisterMessage(SegmentTag)

Segment = _reflection.GeneratedProtocolMessageType('Segment', (_message.Message,), {
  'DESCRIPTOR' : _SEGMENT,
  '__module__' : 'ludwig.profiling.proto.whylogs_messages_pb2'
  # @@protoc_insertion_point(class_scope:dataset_profile.Segment)
  })
_sym_db.RegisterMessage(Segment)

DatasetSegmentHeader = _reflection.GeneratedProtocolMessageType('DatasetSegmentHeader', (_message.Message,), {

  'OffsetsEntry' : _reflection.GeneratedProtocolMessageType('OffsetsEntry', (_message.Message,), {
    'DESCRIPTOR' : _DATASETSEGMENTHEADER_OFFSETSENTRY,
    '__module__' : 'ludwig.profiling.proto.whylogs_messages_pb2'
    # @@protoc_insertion_point(class_scope:dataset_profile.DatasetSegmentHeader.OffsetsEntry)
    })
  ,
  'DESCRIPTOR' : _DATASETSEGMENTHEADER,
  '__module__' : 'ludwig.profiling.proto.whylogs_messages_pb2'
  # @@protoc_insertion_point(class_scope:dataset_profile.DatasetSegmentHeader)
  })
_sym_db.RegisterMessage(DatasetSegmentHeader)
_sym_db.RegisterMessage(DatasetSegmentHeader.OffsetsEntry)


DESCRIPTOR._options = None
_METRICMESSAGE_METRICCOMPONENTSENTRY._options = None
_COLUMNMESSAGE_METRICCOMPONENTSENTRY._options = None
_DATASETPROPERTIES_TAGSENTRY._options = None
_DATASETPROPERTIES_METADATAENTRY._options = None
_CHUNKMESSAGE_METRICCOMPONENTSENTRY._options = None
_DATASETPROFILEHEADER_COLUMNOFFSETSENTRY._options = None
_DATASETPROFILEHEADER_INDEXEDMETRICPATHSENTRY._options = None
_DATASETSEGMENTHEADER_OFFSETSENTRY._options = None
# @@protoc_insertion_point(module_scope)



from dataclasses import Field
from typing import Optional

from ludwig.schema import utils as schema_utils
from ludwig.schema.metadata import COMMON_METADATA
from ludwig.schema.metadata.parameter_metadata import ParameterMetadata
from ludwig.utils.torch_utils import initializer_registry


def DropoutField(default: float = 0.0, description: str = None, parameter_metadata: ParameterMetadata = None) -> Field:
    description = description or "Default dropout rate applied to fully connected layers."
    full_description = description + (
        " Increasing dropout is a common form of regularization to combat overfitting. "
        "The dropout is expressed as the probability of an element to be zeroed out (0.0 means no dropout)."
    )
    parameter_metadata = parameter_metadata or COMMON_METADATA["dropout"]
    return schema_utils.FloatRange(
        default=default,
        min=0,
        max=1,
        description=full_description,
        parameter_metadata=parameter_metadata,
    )


def ResidualField(
    default: bool = False, description: str = None, parameter_metadata: ParameterMetadata = None
) -> Field:
    description = description or (
        "Whether to add a residual connection to each fully connected layer block. "
        "Requires all fully connected layers to have the same `output_size`."
    )
    parameter_metadata = parameter_metadata or COMMON_METADATA["residual"]
    return schema_utils.Boolean(
        default=False,
        description=description,
        parameter_metadata=parameter_metadata,
    )


def NumFCLayersField(default: int = 0, description: str = None, parameter_metadata: ParameterMetadata = None) -> Field:
    description = description or "Number of stacked fully connected layers to apply."
    full_description = description + (
        " Increasing layers adds capacity to the model, enabling it to learn more complex feature interactions."
    )
    parameter_metadata = parameter_metadata or COMMON_METADATA["num_fc_layers"]
    return schema_utils.NonNegativeInteger(
        default=default,
        allow_none=False,
        description=full_description,
        parameter_metadata=parameter_metadata,
    )


def NormField(
    default: Optional[str] = None, description: str = None, parameter_metadata: ParameterMetadata = None
) -> Field:
    description = description or "Default normalization applied at the beginnging of fully connected layers."
    parameter_metadata = parameter_metadata or COMMON_METADATA["norm"]
    return schema_utils.StringOptions(
        ["batch", "layer", "ghost"],
        default=default,
        allow_none=True,
        description=description,
        parameter_metadata=parameter_metadata,
    )


def NormParamsField(description: str = None, parameter_metadata: ParameterMetadata = None) -> Field:
    description = description or "Default parameters passed to the `norm` module."
    parameter_metadata = parameter_metadata or COMMON_METADATA["norm_params"]
    return schema_utils.Dict(
        description=description,
        parameter_metadata=parameter_metadata,
    )


def FCLayersField(description: str = None, parameter_metadata: ParameterMetadata = None) -> Field:
    description = description or (
        "List of dictionaries containing the parameters of all the fully connected layers. "
        "The length of the list determines the number of stacked fully connected layers "
        "and the content of each dictionary determines the parameters for a specific layer. "
        "The available parameters for each layer are: `activation`, `dropout`, `norm`, `norm_params`, "
        "`output_size`, `use_bias`, `bias_initializer` and `weights_initializer`. If any of those values "
        "is missing from the dictionary, the default one provided as a standalone parameter will be used instead."
    )
    parameter_metadata = parameter_metadata or COMMON_METADATA["fc_layers"]
    return schema_utils.DictList(
        description=description,
        parameter_metadata=parameter_metadata,
    )


INITIALIZER_SUFFIX = """
Alternatively it is possible to specify a dictionary with a key `type` that identifies the type of initializer and
other keys for its parameters, e.g. `{type: normal, mean: 0, stddev: 0}`. For a description of the parameters of each
initializer, see [torch.nn.init](https://pytorch.org/docs/stable/nn.init.html).
"""


def BiasInitializerField(
    default: str = "zeros", description: str = None, parameter_metadata: ParameterMetadata = None
) -> Field:
    initializers_str = ", ".join([f"`{i}`" for i in initializer_registry.keys()])
    description = description or "Initializer for the bias vector."
    full_description = f"{description} Options: {initializers_str}. {INITIALIZER_SUFFIX}"
    parameter_metadata = parameter_metadata or COMMON_METADATA["bias_initializer"]
    return schema_utils.InitializerOrDict(
        default=default,
        description=full_description,
        parameter_metadata=parameter_metadata,
    )


def WeightsInitializerField(
    default: str = "xavier_uniform", description: str = None, parameter_metadata: ParameterMetadata = None
) -> Field:
    initializers_str = ", ".join([f"`{i}`" for i in initializer_registry.keys()])
    description = description or "Initializer for the weight matrix."
    full_description = f"{description} Options: {initializers_str}. {INITIALIZER_SUFFIX}"
    parameter_metadata = parameter_metadata or COMMON_METADATA["weights_initializer"]
    return schema_utils.InitializerOrDict(
        default=default,
        description=full_description,
        parameter_metadata=parameter_metadata,
    )


def EmbeddingInitializerField(
    default: Optional[str] = None, description: str = None, parameter_metadata: ParameterMetadata = None
) -> Field:
    initializers_str = ", ".join([f"`{i}`" for i in initializer_registry.keys()])
    description = description or "Initializer for the embedding matrix."
    full_description = f"{description} Options: {initializers_str}."
    parameter_metadata = parameter_metadata or COMMON_METADATA["embedding_initializer"]
    return schema_utils.StringOptions(
        list(initializer_registry.keys()),
        default=default,
        allow_none=True,
        description=full_description,
        parameter_metadata=parameter_metadata,
    )


def EmbeddingSizeField(
    default: int = 256, description: str = None, parameter_metadata: ParameterMetadata = None
) -> Field:
    description = description or (
        "The maximum embedding size. The actual size will be `min(vocabulary_size, embedding_size)` for "
        "`dense` representations and exactly `vocabulary_size` for the `sparse` encoding, where `vocabulary_size` "
        "is the number of unique strings appearing in the training set input column plus the number of "
        "special tokens (`<UNK>`, `<PAD>`, `<SOS>`, `<EOS>`)."
    )
    parameter_metadata = parameter_metadata or COMMON_METADATA["embedding_size"]
    return schema_utils.PositiveInteger(
        default=default,
        description=description,
        parameter_metadata=parameter_metadata,
    )


def EmbeddingsOnCPUField(
    default: bool = False, description: str = None, parameter_metadata: ParameterMetadata = None
) -> Field:
    description = description or (
        "Whether to force the placement of the embedding matrix in regular memory and have the CPU resolve them. "
        "By default embedding matrices are stored on GPU memory if a GPU is used, as it allows for faster access, "
        "but in some cases the embedding matrix may be too large. This parameter forces the placement of the "
        "embedding matrix in regular memory and the CPU is used for embedding lookup, slightly slowing down the "
        "process as a result of data transfer between CPU and GPU memory."
    )
    parameter_metadata = parameter_metadata or COMMON_METADATA["embeddings_on_cpu"]
    return schema_utils.Boolean(
        default=default,
        description=description,
        parameter_metadata=parameter_metadata,
    )


def EmbeddingsTrainableField(
    default: bool = True, description: str = None, parameter_metadata: ParameterMetadata = None
) -> Field:
    description = description or (
        "If `true` embeddings are trained during the training process, if `false` embeddings are fixed. "
        "It may be useful when loading pretrained embeddings for avoiding finetuning them. This parameter "
        "has effect only when `representation` is `dense`; `sparse` one-hot encodings are not trainable."
    )
    parameter_metadata = parameter_metadata or COMMON_METADATA["embeddings_trainable"]
    return schema_utils.Boolean(
        default=default,
        description=description,
        parameter_metadata=parameter_metadata,
    )


def PretrainedEmbeddingsField(
    default: Optional[str] = None, description: str = None, parameter_metadata: ParameterMetadata = None
) -> Field:
    description = description or (
        "Path to a file containing pretrained embeddings. By default `dense` embeddings are initialized "
        "randomly, but this parameter allows to specify a path to a file containing embeddings in the "
        "[GloVe format](https://nlp.stanford.edu/projects/glove/). When the file containing the embeddings is "
        "loaded, only the embeddings with labels present in the vocabulary are kept, the others are discarded. "
        "If the vocabulary contains strings that have no match in the embeddings file, their embeddings are "
        "initialized with the average of all other embedding plus some random noise to make them different "
        "from each other. This parameter has effect only if `representation` is `dense`."
    )
    parameter_metadata = parameter_metadata or COMMON_METADATA["pretrained_embeddings"]
    return schema_utils.String(
        default=default,
        allow_none=True,
        description=description,
        parameter_metadata=parameter_metadata,
    )


def MaxSequenceLengthField(
    default: Optional[int] = None, description: str = None, parameter_metadata: ParameterMetadata = None
) -> Field:
    description = description or "[internal] Maximum sequence length from preprocessing."
    parameter_metadata = parameter_metadata or COMMON_METADATA["max_sequence_length"]
    return schema_utils.PositiveInteger(
        default=default,
        allow_none=True,
        description=description,
        parameter_metadata=parameter_metadata,
    )


def VocabField(
    default: Optional[list] = None, description: str = None, parameter_metadata: ParameterMetadata = None
) -> Field:
    description = description or "[internal] Vocabulary for the encoder from preprocessing."
    parameter_metadata = parameter_metadata or COMMON_METADATA["vocab"]
    return schema_utils.List(
        default=default,
        description=description,
        parameter_metadata=parameter_metadata,
    )


def VocabSizeField(
    default: Optional[list] = None, description: str = None, parameter_metadata: ParameterMetadata = None
) -> Field:
    description = description or "[internal] Size of the vocabulary from preprocessing."
    parameter_metadata = parameter_metadata or COMMON_METADATA["vocab_size"]
    return schema_utils.PositiveInteger(
        default=default,
        allow_none=True,
        description=description,
        parameter_metadata=parameter_metadata,
    )


def RepresentationField(
    default: str = "dense", description: str = None, parameter_metadata: ParameterMetadata = None
) -> Field:
    description = description or (
        "Representation of the embedding. `dense` means the embeddings are initialized randomly, "
        "`sparse` means they are initialized to be one-hot encodings."
    )
    parameter_metadata = parameter_metadata or COMMON_METADATA["representation"]
    return schema_utils.StringOptions(
        ["dense", "sparse"],
        default=default,
        description=description,
        parameter_metadata=parameter_metadata,
    )


def ReduceOutputField(
    default: Optional[str] = "sum", description: str = None, parameter_metadata: ParameterMetadata = None
) -> Field:
    description = description or (
        "How to reduce the output tensor along the `s` sequence length dimension if the rank of the "
        "tensor is greater than 2."
    )
    parameter_metadata = parameter_metadata or COMMON_METADATA["reduce_output"]
    return schema_utils.ReductionOptions(
        default=default,
        description=description,
        parameter_metadata=parameter_metadata,
    )



#! /usr/bin/env python
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging
import math

import torch

from ludwig.api_annotations import DeveloperAPI
from ludwig.data.batcher.base import Batcher

logger = logging.getLogger(__name__)


@DeveloperAPI
class RandomAccessBatcher(Batcher):
    def __init__(self, dataset, sampler, batch_size=128, ignore_last=False, augmentation_pipeline=None):
        # store our dataset as well
        self.dataset = dataset
        self.sampler = sampler
        self.sample_it = iter(self.sampler)

        self.ignore_last = ignore_last
        self.batch_size = batch_size
        self.total_size = len(sampler)
        self.augmentation_pipeline = augmentation_pipeline
        self.steps_per_epoch = self._compute_steps_per_epoch()
        self.index = 0
        self.step = 0

    def next_batch(self):
        if self.last_batch():
            raise StopIteration()

        indices = []
        for _ in range(self.batch_size):
            try:
                indices.append(next(self.sample_it))
                self.index += 1
            except StopIteration:
                break

        sub_batch = {feature_name: self.dataset.get(feature_name, indices) for feature_name in self.dataset.features}

        if self.augmentation_pipeline:
            for feature_name, augmentations in self.augmentation_pipeline.items():
                logger.debug(f"RandomAccessBatcher applying augmentation pipeline to batch for feature {feature_name}")
                sub_batch[feature_name] = augmentations(torch.tensor(sub_batch[feature_name]))

        self.step += 1
        return sub_batch

    def last_batch(self):
        # If our current index in the dataset exceeds the size of the dataset,
        # we've finished the epoch and can indicate that this is the last batch
        if self.index >= self.total_size:
            return True
        # This avoids the case where batch size > total size and no steps have been done.
        # For e.g., batch size = 128 but the dataset only has 100 rows.
        elif self.ignore_last and self.step:
            # index += batch_size after each epoch. So, if our current index in total dataset is 1 less than the total
            # dataset size, then the last batch will only have 1 row. Drop it if this happens.
            if self.index - self.total_size == -1:
                logger.info("Last batch in epoch only has 1 sample and will be dropped.")
                return True
        return False

    def set_epoch(self, epoch, batch_size):
        self.batch_size = batch_size
        self.steps_per_epoch = self._compute_steps_per_epoch()
        self.index = 0
        self.step = 0
        self.sampler.set_epoch(epoch)
        self.sample_it = iter(self.sampler)

    def _compute_steps_per_epoch(self):
        return int(math.ceil(self.total_size / self.batch_size))



import numpy as np

from ludwig.utils.server_utils import NumpyJSONResponse


def test_numpy_json_response():
    response = NumpyJSONResponse({"message": "Ludwig server is up"})

    # Test Python builtin data type encoding.
    assert response.render(None) == b"null"
    assert response.render({}) == b"{}"
    assert response.render(1) == b"1"
    assert response.render(1.0) == b"1.0"
    assert response.render("a") == b'"a"'
    assert response.render([0, 1, 2, 3, 4]) == b"[0,1,2,3,4]"
    assert response.render((0, 1, 2, 3, 4)) == b"[0,1,2,3,4]"
    assert response.render({0, 1, 2, 3, 4}) == b"[0,1,2,3,4]"
    assert response.render({"a": "b"}) == b'{"a":"b"}'

    # Test numpy data type encoding
    for dtype in [np.byte, np.ubyte, np.short, np.ushort, np.int, np.uint, np.longlong, np.ulonglong]:
        x = np.arange(5, dtype=dtype)
        assert response.render(x) == b"[0,1,2,3,4]"
        for i in x:
            assert response.render(i) == f"{i}".encode()

    for dtype in [np.half, np.single, np.double, np.longdouble]:
        x = np.arange(5, dtype=dtype)
        assert response.render(x) == b"[0.0,1.0,2.0,3.0,4.0]"
        for i in x:
            assert response.render(i) == f"{i}".encode()



# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging
from typing import Dict, List, Optional, Tuple, Union

import torch
from torch import nn

from ludwig.constants import TYPE
from ludwig.modules.initializer_modules import get_initializer
from ludwig.utils.data_utils import load_pretrained_embeddings
from ludwig.utils.torch_utils import get_torch_device, LudwigModule

logger = logging.getLogger(__name__)

DEVICE = get_torch_device()


def embedding_matrix(
    vocab: List[str],
    embedding_size: int,
    representation: str = "dense",
    embeddings_trainable: bool = True,
    pretrained_embeddings: Optional[str] = None,
    force_embedding_size: bool = False,
    embedding_initializer: Optional[Union[str, Dict]] = None,
) -> Tuple[nn.Module, int]:
    """Returns initialized torch.nn.Embedding module and embedding size."""

    vocab_size = len(vocab)
    if representation == "dense":
        if pretrained_embeddings:
            embeddings_matrix = load_pretrained_embeddings(pretrained_embeddings, vocab)
            if embeddings_matrix.shape[-1] != embedding_size:
                if not force_embedding_size:
                    embedding_size = embeddings_matrix.shape[-1]
                    logger.info(f"Setting embedding size to be equal to {embeddings_matrix.shape[-1]}.")
                else:
                    raise ValueError(
                        f"The size of the pretrained embeddings is "
                        f"{embeddings_matrix.shape[-1]}, but the specified "
                        f"embedding_size is {embedding_size}. Please change "
                        f"the embedding_size accordingly."
                    )
            embedding_initializer_obj = torch.tensor(embeddings_matrix, dtype=torch.float32)

        else:
            if vocab_size < embedding_size and not force_embedding_size:
                logger.info(
                    f"  embedding_size ({embedding_size}) is greater than "
                    f"vocab_size ({vocab_size}). Setting embedding size to be "
                    f"equal to vocab_size."
                )
                embedding_size = vocab_size

            if embedding_initializer is not None:
                embedding_initializer_obj_ref = get_initializer(embedding_initializer)
            else:
                embedding_initializer_obj_ref = get_initializer({TYPE: "uniform", "a": -1.0, "b": 1.0})
            embedding_initializer_obj = embedding_initializer_obj_ref([vocab_size, embedding_size])

        embeddings = embedding_initializer_obj

    elif representation == "sparse":
        embedding_size = vocab_size
        embeddings = get_initializer("identity")([vocab_size, embedding_size])
        embeddings.requires_grad = False
    else:
        raise Exception(f"Embedding representation {representation} not supported.")

    embeddings = nn.Embedding.from_pretrained(embeddings, freeze=not embeddings_trainable)
    return embeddings, embedding_size


def embedding_matrix_on_device(
    vocab: List[str],
    embedding_size: int,
    representation: str = "dense",
    embeddings_trainable: bool = True,
    pretrained_embeddings: Optional[str] = None,
    force_embedding_size: bool = False,
    embeddings_on_cpu: bool = False,
    embedding_initializer: Optional[str] = None,
) -> Tuple[nn.Module, int]:
    embeddings, embedding_size = embedding_matrix(
        vocab,
        embedding_size,
        representation=representation,
        embeddings_trainable=embeddings_trainable,
        pretrained_embeddings=pretrained_embeddings,
        force_embedding_size=force_embedding_size,
        embedding_initializer=embedding_initializer,
    )
    if embeddings_on_cpu:
        embeddings.to("cpu")
    elif not embeddings_on_cpu and torch.cuda.is_available():
        embeddings.to(device="cuda")

    return embeddings, embedding_size


class Embed(LudwigModule):
    """Module to embed Category, Date, and H3 data types."""

    def __init__(
        self,
        vocab: List[str],
        embedding_size: int,
        representation: str = "dense",
        embeddings_trainable: bool = True,
        pretrained_embeddings: Optional[str] = None,
        force_embedding_size: bool = False,
        embeddings_on_cpu: bool = False,
        dropout: float = 0.0,
        embedding_initializer: Optional[Union[str, Dict]] = None,
    ):
        super().__init__()
        self.supports_masking = True

        self.vocab_size = len(vocab)
        self.embeddings, self.embedding_size = embedding_matrix_on_device(
            vocab,
            embedding_size,
            representation=representation,
            embeddings_trainable=embeddings_trainable,
            pretrained_embeddings=pretrained_embeddings,
            force_embedding_size=force_embedding_size,
            embeddings_on_cpu=embeddings_on_cpu,
            embedding_initializer=embedding_initializer,
        )

        if dropout > 0:
            self.dropout = torch.nn.Dropout(p=dropout)
        else:
            self.dropout = None

    def forward(self, inputs: torch.Tensor, mask: Optional[torch.Tensor] = None) -> torch.Tensor:
        if inputs.ndim != 2 or inputs.shape[1] != 1:
            raise RuntimeError(
                f"Embed only takes inputs of shape [batch x 1]. Received inputs with size: {inputs.size()}"
            )
        embedded = self.embeddings(inputs.long())
        embedded = torch.squeeze(embedded, dim=1)
        if self.dropout:
            embedded = self.dropout(embedded)
        return embedded

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([1])

    @property
    def output_shape(self) -> torch.Size:
        return torch.Size([self.embedding_size])


class EmbedSet(LudwigModule):
    """Module to embed Set data types, works on multi-hot encoded input."""

    def __init__(
        self,
        vocab: List[str],
        embedding_size: int,
        representation: str = "dense",
        embeddings_trainable: bool = True,
        pretrained_embeddings: Optional[str] = None,
        force_embedding_size: bool = False,
        embeddings_on_cpu: bool = False,
        dropout: float = 0.0,
        embedding_initializer: Optional[Union[str, Dict]] = None,
        aggregation_function: str = "sum",
    ):
        super().__init__()
        self.supports_masking = True

        self.vocab_size = len(vocab)
        self.embeddings, self.embedding_size = embedding_matrix_on_device(
            vocab,
            embedding_size,
            representation=representation,
            embeddings_trainable=embeddings_trainable,
            pretrained_embeddings=pretrained_embeddings,
            force_embedding_size=force_embedding_size,
            embeddings_on_cpu=embeddings_on_cpu,
            embedding_initializer=embedding_initializer,
        )

        if dropout > 0:
            self.dropout = torch.nn.Dropout(p=dropout)
        else:
            self.dropout = None

        if aggregation_function == "sum":
            self.aggregation_function = torch.sum
        elif aggregation_function == "avg":
            self.aggregation_function = torch.mean
        else:
            raise ValueError(f"Unsupported aggregation function {aggregation_function}")

        self.register_buffer("vocab_indices", torch.arange(self.vocab_size))

    def forward(self, inputs: torch.Tensor, mask: Optional[torch.Tensor] = None) -> torch.Tensor:
        """
        Params:
            inputs: Boolean multi-hot tensor of size [batch x vocab_size], where
                    inputs[b, i] indicates that token i is present in sample b.
        """
        # Convert multi-hot input to input of indices
        inputs = inputs.int() * self.vocab_indices
        embedded = self.embeddings(inputs.long())
        # Mask out the 0th embedding
        mask = torch.unsqueeze(inputs, -1)
        embedded = embedded * mask
        # Sum over all positive tokens
        embedded = self.aggregation_function(embedded, dim=1)
        if self.dropout:
            embedded = self.dropout(embedded)
        return embedded

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([self.vocab_size])

    @property
    def output_shape(self) -> torch.Size:
        return torch.Size([self.embedding_size])

    @property
    def input_dtype(self):
        return torch.bool


class EmbedWeighted(LudwigModule):
    """Module to embed Bag data type, works on input of token frequencies."""

    def __init__(
        self,
        vocab: List[str],
        embedding_size: int,
        representation: str = "dense",
        embeddings_trainable: bool = True,
        pretrained_embeddings: Optional[str] = None,
        force_embedding_size: bool = False,
        embeddings_on_cpu: bool = False,
        dropout: float = 0.0,
        embedding_initializer: Optional[str] = None,
    ):
        super().__init__()

        self.embeddings, self.embedding_size = embedding_matrix_on_device(
            vocab,
            embedding_size,
            representation=representation,
            embeddings_trainable=embeddings_trainable,
            pretrained_embeddings=pretrained_embeddings,
            force_embedding_size=force_embedding_size,
            embeddings_on_cpu=embeddings_on_cpu,
            embedding_initializer=embedding_initializer,
        )
        self.vocab_size = len(vocab)

        if dropout > 0:
            self.dropout = nn.Dropout(dropout)
        else:
            self.dropout = None

        self.register_buffer("vocab_indices", torch.arange(self.vocab_size, dtype=torch.int32))

    def forward(self, inputs: torch.Tensor, mask: Optional[torch.Tensor] = None) -> torch.Tensor:
        """
        Params:
            inputs: Tensor of frequencies, where inputs[b, i] represents
                    frequency of token i in sample b of batch.
        """
        # Convert to multi-hot input
        signed_input = (inputs != 0).type(torch.int32)
        multiple_hot_indexes = signed_input * self.vocab_indices
        embedded = self.embeddings(multiple_hot_indexes)
        # Mask out the 0th embedding
        mask = torch.unsqueeze(inputs, -1)
        weighted_embedded = embedded * mask
        # Sum over the all the positive indices
        embedded_reduced = torch.sum(weighted_embedded, dim=1)
        if self.dropout:
            embedded_reduced = self.dropout(embedded_reduced)
        return embedded_reduced

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([self.vocab_size])

    @property
    def output_shape(self) -> torch.Size:
        return torch.Size([self.embedding_size])


# TODO(shreya): Implement sparse embedding lookup.
# class EmbedSparse(LudwigModule):
#     def __init__(
#             self,
#             vocab: List[str],
#             embedding_size: int = 50,
#             representation: str = 'dense',
#             embeddings_trainable: bool = True,
#             pretrained_embeddings: Optional[str] = None,
#             force_embedding_size: bool = False,
#             embeddings_on_cpu: bool = False,
#             dropout: float = 0.0,
#             embedding_initializer: Optional[str] = None,
#             reduce_output: str = 'sum'
#     ):
#         super().__init__()

#         self.embeddings, self.embedding_size = embedding_matrix_on_device(
#             vocab,
#             embedding_size,
#             representation=representation,
#             embeddings_trainable=embeddings_trainable,
#             pretrained_embeddings=pretrained_embeddings,
#             force_embedding_size=force_embedding_size,
#             embeddings_on_cpu=embeddings_on_cpu,
#             embedding_initializer=embedding_initializer,
#         )

#         if dropout > 0:
#             self.dropout = nn.Dropout(dropout)
#         else:
#             self.dropout = None

#         self.reduce_output = reduce_output

#     def forward(self, inputs: torch.Tensor):
#         # idx = tf.where(tf.equal(inputs, True))
#         # TODO(shreya): Check if this is equivalent
#         idx = torch.nonzero(inputs)

#         # sparse_multiple_hot_indexes = tf.SparseTensor(
#         #     idx,
#         #     idx[:, 1],
#         #     tf.shape(inputs, out_type=tf.int64)
#         # )
#         sparse_multiple_hot_index = torch.sparse_coo_tensor(
#             idx, idx[:, 1], inputs.shape
#         )

#         # TODO(shreya): Check if supported in torch
#         # embedded_reduced = tf.nn.embedding_lookup_sparse(
#         #     self.embeddings,
#         #     sparse_multiple_hot_indexes,
#         #     sp_weights=None,
#         #     combiner=self.reduce_output
#         # )

#         # if self.dropout:
#         #     embedded_reduced = self.dropout(embedded_reduced)

#         # return embedded_reduced
#         return None


class EmbedSequence(LudwigModule):
    def __init__(
        self,
        vocab: List[str],
        embedding_size: int,
        max_sequence_length: int,
        representation: str = "dense",
        embeddings_trainable: bool = True,
        pretrained_embeddings: Optional[str] = None,
        force_embedding_size: bool = False,
        embeddings_on_cpu: bool = False,
        dropout: float = 0.0,
        embedding_initializer: Optional[str] = None,
    ):
        super().__init__()
        self.supports_masking = True

        self.vocab_size = len(vocab)
        self.max_sequence_length = max_sequence_length
        self.embeddings, self.embedding_size = embedding_matrix_on_device(
            vocab,
            embedding_size,
            representation=representation,
            embeddings_trainable=embeddings_trainable,
            pretrained_embeddings=pretrained_embeddings,
            force_embedding_size=force_embedding_size,
            embeddings_on_cpu=embeddings_on_cpu,
            embedding_initializer=embedding_initializer,
        )

        if dropout > 0:
            self.dropout = nn.Dropout(dropout)
        else:
            self.dropout = None

    def forward(self, inputs: torch.Tensor, mask: Optional[torch.Tensor] = None):
        if inputs.dtype not in [torch.int, torch.long]:
            raise RuntimeError(
                f"Expected tensor of type torch.int or torch.long as input." f"Received {inputs.dtype} instead."
            )

        embedded = self.embeddings(inputs)
        if self.dropout:
            embedded = self.dropout(embedded)
        return embedded

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([self.max_sequence_length])

    @property
    def output_shape(self) -> torch.Size:
        return torch.Size([self.max_sequence_length, self.embedding_size])


class TokenAndPositionEmbedding(LudwigModule):
    def __init__(
        self,
        max_sequence_length,
        vocab,
        embedding_size,
        representation="dense",
        embeddings_trainable=True,
        pretrained_embeddings=None,
        force_embedding_size=False,
        embeddings_on_cpu=False,
        dropout=0.0,
        embedding_initializer=None,
    ):
        super().__init__()
        self.max_sequence_length = max_sequence_length
        self.embedding_size = embedding_size
        self.token_embed = EmbedSequence(
            vocab=vocab,
            embedding_size=embedding_size,
            max_sequence_length=max_sequence_length,
            representation=representation,
            embeddings_trainable=embeddings_trainable,
            pretrained_embeddings=pretrained_embeddings,
            force_embedding_size=force_embedding_size,
            embeddings_on_cpu=embeddings_on_cpu,
            dropout=dropout,
            embedding_initializer=embedding_initializer,
        )
        self.position_embed = nn.Embedding(
            num_embeddings=max_sequence_length, embedding_dim=self.token_embed.embedding_size
        )
        self.register_buffer("positions", torch.arange(0, max_sequence_length))

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([self.max_sequence_length])

    @property
    def output_shape(self) -> torch.Size:
        return self.token_embed.output_shape

    def forward(self, inputs, mask: Optional[torch.Tensor] = None):
        positions_hidden = self.position_embed(self.positions)
        token_hidden = self.token_embed(inputs)
        return token_hidden + positions_hidden



# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import json
import os
import os.path
import uuid
from typing import Dict, Tuple

import pytest

from ludwig.backend import initialize_backend
from ludwig.constants import (
    ACCURACY,
    AUTO,
    BATCH_SIZE,
    CATEGORY,
    COMBINER,
    EXECUTOR,
    HYPEROPT,
    INPUT_FEATURES,
    MAX_CONCURRENT_TRIALS,
    MODEL_ECD,
    MODEL_GBM,
    MODEL_TYPE,
    NAME,
    OUTPUT_FEATURES,
    RAY,
    TEXT,
    TRAINER,
    TYPE,
)
from ludwig.globals import HYPEROPT_STATISTICS_FILE_NAME
from ludwig.hyperopt.results import HyperoptResults
from ludwig.hyperopt.run import hyperopt
from ludwig.hyperopt.utils import update_hyperopt_params_with_defaults
from ludwig.schema.model_config import ModelConfig
from ludwig.utils import fs_utils
from ludwig.utils.data_utils import load_json, use_credentials
from tests.integration_tests.utils import (
    category_feature,
    generate_data,
    minio_test_creds,
    private_param,
    remote_tmpdir,
    text_feature,
)

ray = pytest.importorskip("ray")

from ludwig.hyperopt.execution import get_build_hyperopt_executor, RayTuneExecutor  # noqa

pytestmark = pytest.mark.distributed

RANDOM_SEARCH_SIZE = 2

HYPEROPT_CONFIG = {
    "parameters": {
        # using only float parameter as common in all search algorithms
        "trainer.learning_rate": {"space": "loguniform", "lower": 0.001, "upper": 0.1},
    },
    "goal": "minimize",
    "executor": {TYPE: "ray", "num_samples": 2, "scheduler": {TYPE: "fifo"}},
    "search_alg": {TYPE: "variant_generator"},
}

SEARCH_ALGS_FOR_TESTING = [
    # None,
    # "variant_generator",
    "random",
    "bohb",
    # "hyperopt",
    # "ax",
    # "bayesopt",
    # "blendsearch",
    # "cfo",
    # "dragonfly",
    # "hebo",
    # "skopt",
    # "optuna",
]

SCHEDULERS_FOR_TESTING = [
    "fifo",
    "asynchyperband",
    # "async_hyperband",
    # "median_stopping_rule",
    # "medianstopping",
    # "hyperband",
    # "hb_bohb",
    # "pbt",
    # "pb2",  commented out for now: https://github.com/ray-project/ray/issues/24815
    # "resource_changing",
]


def _setup_ludwig_config(dataset_fp: str, model_type: str = MODEL_ECD) -> Tuple[Dict, str]:
    input_features = [category_feature(encoder={"vocab_size": 3})]
    output_features = [category_feature(decoder={"vocab_size": 3})]

    rel_path = generate_data(input_features, output_features, dataset_fp)

    trainer_cfg = {"learning_rate": 0.001}
    if model_type == MODEL_ECD:
        trainer_cfg["epochs"] = 2
    else:
        trainer_cfg["num_boost_round"] = 2
        # Disable feature filtering to avoid having no features due to small test dataset,
        # see https://stackoverflow.com/a/66405983/5222402
        trainer_cfg["feature_pre_filter"] = False

    config = {
        MODEL_TYPE: model_type,
        INPUT_FEATURES: input_features,
        OUTPUT_FEATURES: output_features,
        COMBINER: {TYPE: "concat"},
        TRAINER: trainer_cfg,
    }

    config = ModelConfig.from_dict(config).to_dict()

    return config, rel_path


@pytest.mark.parametrize("search_alg", SEARCH_ALGS_FOR_TESTING)
@pytest.mark.parametrize("model_type", [MODEL_ECD, MODEL_GBM])
def test_hyperopt_search_alg(
    search_alg,
    model_type,
    csv_filename,
    tmpdir,
    ray_cluster_7cpu,
    validate_output_feature=False,
    validation_metric=None,
):
    config, rel_path = _setup_ludwig_config(csv_filename, model_type)

    hyperopt_config = HYPEROPT_CONFIG.copy()

    # finalize hyperopt config settings
    if search_alg == "dragonfly":
        hyperopt_config["search_alg"] = {
            TYPE: search_alg,
            "domain": "euclidean",
            "optimizer": "random",
        }
    elif search_alg is None:
        hyperopt_config["search_alg"] = {}
    else:
        hyperopt_config["search_alg"] = {
            TYPE: search_alg,
        }

    if validate_output_feature:
        hyperopt_config["output_feature"] = config[OUTPUT_FEATURES][0][NAME]
    if validation_metric:
        hyperopt_config["validation_metric"] = validation_metric

    update_hyperopt_params_with_defaults(hyperopt_config)

    backend = initialize_backend("local")
    if hyperopt_config[EXECUTOR].get(MAX_CONCURRENT_TRIALS) == AUTO:
        hyperopt_config[EXECUTOR][MAX_CONCURRENT_TRIALS] = backend.max_concurrent_trials(hyperopt_config)

    parameters = hyperopt_config["parameters"]
    split = hyperopt_config["split"]
    output_feature = hyperopt_config["output_feature"]
    metric = hyperopt_config["metric"]
    goal = hyperopt_config["goal"]
    executor = hyperopt_config["executor"]
    search_alg = hyperopt_config["search_alg"]

    hyperopt_executor = get_build_hyperopt_executor(RAY)(
        parameters, output_feature, metric, goal, split, search_alg=search_alg, **executor
    )
    results = hyperopt_executor.execute(config, dataset=rel_path, output_directory=tmpdir)
    assert isinstance(results, HyperoptResults)

    with hyperopt_executor._get_best_model_path(
        results.experiment_analysis.best_trial, results.experiment_analysis, {}
    ) as path:
        assert path is not None
        assert isinstance(path, str)


@pytest.mark.parametrize("model_type", [MODEL_ECD, MODEL_GBM])
def test_hyperopt_executor_with_metric(model_type, csv_filename, tmpdir, ray_cluster_7cpu):
    test_hyperopt_search_alg(
        "variant_generator",
        model_type,
        csv_filename,
        tmpdir,
        ray_cluster_7cpu,
        validate_output_feature=True,
        validation_metric=ACCURACY,
    )


@pytest.mark.parametrize("scheduler", SCHEDULERS_FOR_TESTING)
@pytest.mark.parametrize("model_type", [MODEL_ECD, MODEL_GBM])
def test_hyperopt_scheduler(
    scheduler, model_type, csv_filename, tmpdir, ray_cluster_7cpu, validate_output_feature=False, validation_metric=None
):
    config, rel_path = _setup_ludwig_config(csv_filename, model_type)

    hyperopt_config = HYPEROPT_CONFIG.copy()

    # finalize hyperopt config settings
    if scheduler == "pb2":
        # setup scheduler hyperparam_bounds parameter
        min = hyperopt_config["parameters"]["trainer.learning_rate"]["lower"]
        max = hyperopt_config["parameters"]["trainer.learning_rate"]["upper"]
        hyperparam_bounds = {
            "trainer.learning_rate": [min, max],
        }
        hyperopt_config["executor"]["scheduler"] = {
            TYPE: scheduler,
            "hyperparam_bounds": hyperparam_bounds,
        }
    else:
        hyperopt_config["executor"]["scheduler"] = {
            TYPE: scheduler,
        }

    if validate_output_feature:
        hyperopt_config["output_feature"] = config[OUTPUT_FEATURES][0][NAME]
    if validation_metric:
        hyperopt_config["validation_metric"] = validation_metric

    backend = initialize_backend("local")
    update_hyperopt_params_with_defaults(hyperopt_config)
    if hyperopt_config[EXECUTOR].get(MAX_CONCURRENT_TRIALS) == AUTO:
        hyperopt_config[EXECUTOR][MAX_CONCURRENT_TRIALS] = backend.max_concurrent_trials(hyperopt_config)

    parameters = hyperopt_config["parameters"]
    split = hyperopt_config["split"]
    output_feature = hyperopt_config["output_feature"]
    metric = hyperopt_config["metric"]
    goal = hyperopt_config["goal"]
    executor = hyperopt_config["executor"]
    search_alg = hyperopt_config["search_alg"]

    # TODO: Determine if we still need this if-then-else construct
    if search_alg[TYPE] in {""}:
        with pytest.raises(ImportError):
            get_build_hyperopt_executor(RAY)(
                parameters, output_feature, metric, goal, split, search_alg=search_alg, **executor
            )
    else:
        hyperopt_executor = get_build_hyperopt_executor(RAY)(
            parameters, output_feature, metric, goal, split, search_alg=search_alg, **executor
        )
        raytune_results = hyperopt_executor.execute(config, dataset=rel_path, output_directory=tmpdir)
        assert isinstance(raytune_results, HyperoptResults)


def _run_hyperopt_run_hyperopt(csv_filename, search_space, tmpdir, backend, ray_cluster_7cpu):
    input_features = [category_feature(encoder={"vocab_size": 3})]
    output_features = [category_feature(decoder={"vocab_size": 3})]

    rel_path = generate_data(input_features, output_features, csv_filename)

    config = {
        INPUT_FEATURES: input_features,
        OUTPUT_FEATURES: output_features,
        COMBINER: {TYPE: "concat"},
        TRAINER: {"epochs": 2, "learning_rate": 0.001, BATCH_SIZE: 128},
        "backend": backend,
    }

    output_feature_name = output_features[0][NAME]

    if search_space == "random":
        # random search will be size of num_samples
        search_parameters = {
            "trainer.learning_rate": {
                "lower": 0.0001,
                "upper": 0.01,
                "space": "loguniform",
            },
            output_feature_name
            + ".fc_layers": {
                "space": "choice",
                "categories": [
                    [{"output_size": 64}, {"output_size": 32}],
                    [{"output_size": 64}],
                    [{"output_size": 32}],
                ],
            },
            output_feature_name + ".output_size": {"space": "choice", "categories": [16, 21, 26, 31, 36]},
        }
    else:
        # grid search space will be product each parameter size
        search_parameters = {
            "trainer.learning_rate": {"space": "grid_search", "values": [0.001, 0.01]},
            output_feature_name + ".output_size": {"space": "grid_search", "values": [16, 21]},
        }

    hyperopt_configs = {
        "parameters": search_parameters,
        "goal": "minimize",
        "output_feature": output_feature_name,
        "validation_metrics": "loss",
        "executor": {
            TYPE: "ray",
            "num_samples": 1 if search_space == "grid" else RANDOM_SEARCH_SIZE,
            "max_concurrent_trials": 1,
        },
        "search_alg": {TYPE: "variant_generator"},
    }

    # add hyperopt parameter space to the config
    config[HYPEROPT] = hyperopt_configs

    experiment_name = f"test_hyperopt_{uuid.uuid4().hex}"
    hyperopt_results = hyperopt(config, dataset=rel_path, output_directory=tmpdir, experiment_name=experiment_name)
    if search_space == "random":
        assert hyperopt_results.experiment_analysis.results_df.shape[0] == RANDOM_SEARCH_SIZE
    else:
        # compute size of search space for grid search
        grid_search_size = 1
        for k, v in search_parameters.items():
            grid_search_size *= len(v["values"])
        assert hyperopt_results.experiment_analysis.results_df.shape[0] == grid_search_size

    # check for return results
    assert isinstance(hyperopt_results, HyperoptResults)

    # check for existence of the hyperopt statistics file
    with use_credentials(minio_test_creds()):
        assert fs_utils.path_exists(os.path.join(tmpdir, experiment_name, HYPEROPT_STATISTICS_FILE_NAME))
        for trial in hyperopt_results.experiment_analysis.trials:
            assert fs_utils.path_exists(
                os.path.join(tmpdir, experiment_name, f"trial_{trial.trial_id}"),
            )

    with RayTuneExecutor._get_best_model_path(
        hyperopt_results.experiment_analysis.best_trial, hyperopt_results.experiment_analysis, minio_test_creds()
    ) as path:
        assert path is not None
        assert isinstance(path, str)
        assert "model" in os.listdir(path)


@pytest.mark.parametrize("search_space", ["random", "grid"])
def test_hyperopt_run_hyperopt(csv_filename, search_space, tmpdir, ray_cluster_7cpu):
    _run_hyperopt_run_hyperopt(csv_filename, search_space, tmpdir, "local", ray_cluster_7cpu)


@pytest.mark.parametrize("fs_protocol,bucket", [private_param(("s3", "ludwig-tests"))], ids=["s3"])
def test_hyperopt_sync_remote(fs_protocol, bucket, csv_filename, ray_cluster_7cpu):
    backend = {
        "type": "local",
        "credentials": {
            "artifacts": minio_test_creds(),
        },
    }

    with remote_tmpdir(fs_protocol, bucket) as tmpdir:
        _run_hyperopt_run_hyperopt(
            csv_filename,
            "random",
            tmpdir,
            backend,
            ray_cluster_7cpu,
        )


def test_hyperopt_with_feature_specific_parameters(csv_filename, tmpdir, ray_cluster_7cpu):
    input_features = [
        text_feature(name="utterance", reduce_output="sum"),
        category_feature(vocab_size=3),
    ]

    output_features = [category_feature(vocab_size=3, output_feature=True)]

    rel_path = generate_data(input_features, output_features, csv_filename)

    filter_size_search_space = [5, 7]
    embedding_size_search_space = [4, 8, 12]

    config = {
        INPUT_FEATURES: input_features,
        OUTPUT_FEATURES: output_features,
        COMBINER: {TYPE: "concat", "num_fc_layers": 2},
        TRAINER: {"epochs": 1, "learning_rate": 0.001, BATCH_SIZE: 128},
        HYPEROPT: {
            "parameters": {
                input_features[0][NAME]
                + ".encoder.filter_size": {"space": "choice", "categories": filter_size_search_space},
                input_features[1][NAME]
                + ".encoder.embedding_size": {"space": "choice", "categories": embedding_size_search_space},
            },
            "goal": "minimize",
            "output_feature": output_features[0][NAME],
            "validation_metrics": "loss",
            "executor": {TYPE: "ray", "num_samples": 1},
            "search_alg": {TYPE: "variant_generator"},
        },
    }

    hyperopt_results = hyperopt(config, dataset=rel_path, output_directory=tmpdir, experiment_name="test_hyperopt")
    hyperopt_results_df = hyperopt_results.experiment_analysis.results_df

    model_parameters = json.load(
        open(
            os.path.join(
                hyperopt_results_df.iloc[0]["trial_dir"], "test_hyperopt_run", "model", "model_hyperparameters.json"
            )
        )
    )

    for input_feature in model_parameters[INPUT_FEATURES]:
        if input_feature[TYPE] == TEXT:
            assert input_feature["encoder"]["filter_size"] in filter_size_search_space
        elif input_feature[TYPE] == CATEGORY:
            assert input_feature["encoder"]["embedding_size"] in embedding_size_search_space


def test_hyperopt_old_config(csv_filename, tmpdir, ray_cluster_7cpu):
    old_config = {
        "ludwig_version": "0.4",
        INPUT_FEATURES: [
            {"name": "cat1", TYPE: "category", "encoder": {"vocab_size": 2}},
            {"name": "num1", TYPE: "number"},
        ],
        OUTPUT_FEATURES: [
            {"name": "bin1", TYPE: "binary"},
        ],
        TRAINER: {"epochs": 2, BATCH_SIZE: 128},
        HYPEROPT: {
            EXECUTOR: {
                TYPE: "ray",
                "time_budget_s": 200,
                "cpu_resources_per_trial": 1,
            },
            "sampler": {
                TYPE: "ray",
                "scheduler": {
                    TYPE: "async_hyperband",
                    "max_t": 200,
                    "time_attr": "time_total_s",
                    "grace_period": 72,
                    "reduction_factor": 5,
                },
                "search_alg": {
                    TYPE: HYPEROPT,
                    "random_state_seed": 42,
                },
                "num_samples": 2,
            },
            "parameters": {
                "trainer.batch_size": {
                    "space": "choice",
                    "categories": [64, 128, 256],
                },
                "trainer.learning_rate": {
                    "space": "loguniform",
                    "lower": 0.001,
                    "upper": 0.1,
                },
            },
        },
    }

    input_features = old_config[INPUT_FEATURES]
    output_features = old_config[OUTPUT_FEATURES]
    rel_path = generate_data(input_features, output_features, csv_filename)

    hyperopt(old_config, dataset=rel_path, output_directory=tmpdir, experiment_name="test_hyperopt")


def test_hyperopt_nested_parameters(csv_filename, tmpdir, ray_cluster_7cpu):
    config = {
        INPUT_FEATURES: [
            {"name": "cat1", TYPE: "category", "encoder": {"vocab_size": 2}},
            {"name": "num1", TYPE: "number"},
        ],
        OUTPUT_FEATURES: [
            {"name": "bin1", TYPE: "binary"},
        ],
        TRAINER: {"epochs": 2, BATCH_SIZE: 128},
        HYPEROPT: {
            EXECUTOR: {
                TYPE: "ray",
                "time_budget_s": 200,
                "cpu_resources_per_trial": 1,
                "num_samples": 2,
                "scheduler": {TYPE: "fifo"},
            },
            "search_alg": {TYPE: "variant_generator"},
            "parameters": {
                ".": {
                    "space": "choice",
                    "categories": [
                        {
                            "combiner": {
                                "type": "tabnet",
                                "bn_virtual_bs": 256,
                            },
                            "trainer": {
                                "learning_rate_scaling": "sqrt",
                                "learning_rate_scheduler": {
                                    "decay": "exponential",
                                    "decay_steps": 20000,
                                    "decay_rate": 0.8,
                                },
                                "optimizer": {"type": "adam"},
                            },
                        },
                        {
                            "combiner": {"type": "concat"},
                            "trainer": {"learning_rate_scaling": "linear"},
                        },
                    ],
                },
                "trainer.learning_rate": {"space": "choice", "categories": [0.7, 0.42]},
            },
        },
    }

    input_features = config[INPUT_FEATURES]
    output_features = config[OUTPUT_FEATURES]
    rel_path = generate_data(input_features, output_features, csv_filename)

    results = hyperopt(
        config,
        dataset=rel_path,
        output_directory=tmpdir,
        experiment_name="test_hyperopt_nested_params",
    )

    results_df = results.experiment_analysis.results_df
    assert len(results_df) == 2

    for _, trial_meta in results_df.iterrows():
        trial_dir = trial_meta["trial_dir"]
        trial_config = load_json(
            os.path.join(trial_dir, "test_hyperopt_nested_params_run", "model", "model_hyperparameters.json")
        )

        assert len(trial_config[INPUT_FEATURES]) == len(config[INPUT_FEATURES])
        assert len(trial_config[OUTPUT_FEATURES]) == len(config[OUTPUT_FEATURES])

        assert trial_config[COMBINER][TYPE] in {"tabnet", "concat"}
        if trial_config[COMBINER][TYPE] == "tabnet":
            assert trial_config[COMBINER]["bn_virtual_bs"] == 256
            assert trial_config[TRAINER]["learning_rate_scaling"] == "sqrt"
            assert trial_config[TRAINER]["learning_rate_scheduler"]["decay"] == "exponential"
            assert trial_config[TRAINER]["learning_rate_scheduler"]["decay_steps"] == 20000
            assert trial_config[TRAINER]["learning_rate_scheduler"]["decay_rate"] == 0.8
            assert trial_config[TRAINER]["optimizer"]["type"] == "adam"
        else:
            assert trial_config[TRAINER]["learning_rate_scaling"] == "linear"

        assert trial_config[TRAINER]["learning_rate"] in {0.7, 0.42}


def test_hyperopt_without_config_defaults(csv_filename, tmpdir, ray_cluster_7cpu):
    input_features = [category_feature(encoder={"vocab_size": 3})]
    output_features = [category_feature(decoder={"vocab_size": 3})]

    rel_path = generate_data(input_features, output_features, csv_filename)

    config = {
        INPUT_FEATURES: input_features,
        OUTPUT_FEATURES: output_features,
        COMBINER: {TYPE: "concat"},
        TRAINER: {"train_steps": 5, "learning_rate": 0.001, BATCH_SIZE: 128},
        # Missing search_alg and executor, but should still work
        HYPEROPT: {
            "parameters": {
                "trainer.learning_rate": {
                    "lower": 0.0001,
                    "upper": 0.01,
                    "space": "loguniform",
                }
            },
            "goal": "minimize",
            "output_feature": output_features[0]["name"],
            "metric": "loss",
        },
    }

    experiment_name = f"test_hyperopt_{uuid.uuid4().hex}"
    hyperopt_results = hyperopt(config, dataset=rel_path, output_directory=tmpdir, experiment_name=experiment_name)
    assert hyperopt_results.experiment_analysis.results_df.shape[0] == 10


def test_hyperopt_with_time_budget(csv_filename, tmpdir, ray_cluster_7cpu):
    """Tests that incomplete checkpoints created by RayTune when time budget is hit doesn't throw errors because of
    missing .tune_metadata files in the checkpoint directories."""
    input_features = [text_feature()]
    output_features = [category_feature(output_feature=True)]

    rel_path = generate_data(input_features, output_features, csv_filename)

    config = {
        INPUT_FEATURES: input_features,
        OUTPUT_FEATURES: output_features,
        COMBINER: {TYPE: "concat"},
        HYPEROPT: {
            "goal": "minimize",
            "metric": "loss",
            "output_feature": output_features[0]["name"],
            "search_alg": {TYPE: "variant_generator"},
            "executor": {
                "type": "ray",
                # Ensure there is enough time for some trials to start and also for some to terminate
                # to reproduce the exact issue of missing .tune_metadata files.
                "time_budget_s": 120,
                "cpu_resources_per_trial": 1,
                "num_samples": 20,
                "scheduler": {TYPE: "fifo"},
            },
            "parameters": {
                "trainer.learning_rate": {
                    "lower": 0.0001,
                    "upper": 0.01,
                    "space": "loguniform",
                }
            },
        },
    }

    experiment_name = f"test_hyperopt_{uuid.uuid4().hex}"
    hyperopt(config, dataset=rel_path, output_directory=tmpdir, experiment_name=experiment_name)



from abc import ABC
from typing import Any, Dict, List, Union

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import BINARY, MODEL_ECD, MODEL_GBM, NUMBER, VECTOR
from ludwig.schema import common_fields
from ludwig.schema import utils as schema_utils
from ludwig.schema.encoders.utils import register_encoder_config
from ludwig.schema.metadata import ENCODER_METADATA
from ludwig.schema.utils import ludwig_dataclass


@DeveloperAPI
@ludwig_dataclass
class BaseEncoderConfig(schema_utils.BaseMarshmallowConfig, ABC):
    """Base class for encoders."""

    type: str

    skip: bool = schema_utils.Boolean(
        False,
        "[internal] Whether to skip encoder and use input as output.",
        parameter_metadata=ENCODER_METADATA["BaseEncoder"]["skip"],
    )

    def get_fixed_preprocessing_params(self) -> Dict[str, Any]:
        return {}

    def is_pretrained(self) -> bool:
        return False

    def can_cache_embeddings(self) -> bool:
        return False


@DeveloperAPI
@register_encoder_config("passthrough", [BINARY, NUMBER, VECTOR], model_types=[MODEL_ECD, MODEL_GBM])
@ludwig_dataclass
class PassthroughEncoderConfig(BaseEncoderConfig):
    """PassthroughEncoderConfig is a dataclass that configures the parameters used for a passthrough encoder."""

    @staticmethod
    def module_name():
        return "PassthroughEncoder"

    type: str = schema_utils.ProtectedString(
        "passthrough",
        description=ENCODER_METADATA["PassthroughEncoder"]["type"].long_description,
    )


@DeveloperAPI
@register_encoder_config("dense", [BINARY, NUMBER, VECTOR])
@ludwig_dataclass
class DenseEncoderConfig(BaseEncoderConfig):
    """DenseEncoderConfig is a dataclass that configures the parameters used for a dense encoder."""

    @staticmethod
    def module_name():
        return "DenseEncoder"

    type: str = schema_utils.ProtectedString(
        "dense",
        description=ENCODER_METADATA["DenseEncoder"]["type"].long_description,
    )

    dropout: float = common_fields.DropoutField()

    activation: str = schema_utils.ActivationOptions()

    input_size: int = schema_utils.PositiveInteger(
        default=None,
        allow_none=True,
        description="Size of the input to the dense encoder.",
        parameter_metadata=ENCODER_METADATA["DenseEncoder"]["input_size"],
    )

    output_size: int = schema_utils.PositiveInteger(
        default=256,
        description="Size of the output of the feature.",
        parameter_metadata=ENCODER_METADATA["DenseEncoder"]["output_size"],
    )

    use_bias: bool = schema_utils.Boolean(
        default=True,
        description="Whether the layer uses a bias vector.",
        parameter_metadata=ENCODER_METADATA["DenseEncoder"]["use_bias"],
    )

    bias_initializer: Union[str, dict] = common_fields.BiasInitializerField()

    weights_initializer: Union[str, dict] = common_fields.WeightsInitializerField()

    norm: str = common_fields.NormField()

    norm_params: dict = common_fields.NormParamsField()

    num_layers: int = common_fields.NumFCLayersField(default=1)

    fc_layers: List[dict] = common_fields.FCLayersField()



from typing import Optional

import pytest
import torch

from ludwig.modules import loss_modules


def from_float(v: float) -> torch.Tensor:
    return torch.tensor(v).float()


@pytest.mark.parametrize("preds", [torch.arange(6).reshape(3, 2).float()])
@pytest.mark.parametrize("target", [torch.arange(6, 12).reshape(3, 2).float()])
@pytest.mark.parametrize("output", [torch.tensor(36).float()])
def test_mse_loss(preds: torch.Tensor, target: torch.Tensor, output: torch.Tensor):
    loss = loss_modules.MSELoss()
    assert loss(preds, target) == output


@pytest.mark.parametrize("preds", [torch.arange(6).reshape(3, 2).float()])
@pytest.mark.parametrize("target", [torch.arange(6, 12).reshape(3, 2).float()])
@pytest.mark.parametrize("output", [torch.tensor(6).float()])
def test_mae_loss(preds: torch.Tensor, target: torch.Tensor, output: torch.Tensor):
    loss = loss_modules.MAELoss()
    assert loss(preds, target) == output


@pytest.mark.parametrize("preds", [torch.arange(6).reshape(3, 2).float()])
@pytest.mark.parametrize("target", [torch.arange(6, 12).reshape(3, 2).float()])
@pytest.mark.parametrize("output", [torch.tensor(6).float()])
def test_rmse_loss(preds: torch.Tensor, target: torch.Tensor, output: torch.Tensor):
    loss = loss_modules.RMSELoss()
    assert loss(preds, target) == output


@pytest.mark.parametrize("preds", [torch.arange(6).reshape(3, 2).float()])
@pytest.mark.parametrize("target", [torch.arange(6, 12).reshape(3, 2).float()])
@pytest.mark.parametrize("output", [torch.tensor(0.7527).float()])
def test_rmspe_loss(preds: torch.Tensor, target: torch.Tensor, output: torch.Tensor):
    loss = loss_modules.RMSPELoss()
    assert torch.isclose(loss(preds, target), output, rtol=0.0001)


@pytest.mark.parametrize("preds", [torch.tensor([[0.1, 0.2]]).float()])
@pytest.mark.parametrize("target", [torch.tensor([[0.0, 0.2]]).float()])
@pytest.mark.parametrize("output", [torch.tensor(707.1068).float()])
def test_rmspe_loss_zero_targets(preds: torch.Tensor, target: torch.Tensor, output: torch.Tensor):
    loss = loss_modules.RMSPELoss()
    assert torch.isclose(loss(preds, target), output, rtol=0.0001)


@pytest.mark.parametrize(
    "confidence_penalty,positive_class_weight,robust_lambda,output",
    [
        (0.0, None, 0, from_float(-21.4655)),
        (2.0, None, 0, from_float(-21.1263)),
        (0.0, 2.0, 0, from_float(-20.1222)),
        (0.0, None, 2, from_float(22.4655)),
        (2, 2, 2, from_float(21.4614)),
    ],
)
@pytest.mark.parametrize("preds", [torch.arange(6).reshape(3, 2).float()])
@pytest.mark.parametrize("target", [torch.arange(6, 12).reshape(3, 2).float()])
def test_bwcew_loss(
    preds: torch.Tensor,
    target: torch.Tensor,
    confidence_penalty: float,
    positive_class_weight: Optional[float],
    robust_lambda: int,
    output: torch.Tensor,
):
    loss = loss_modules.BWCEWLoss(
        positive_class_weight=positive_class_weight, robust_lambda=robust_lambda, confidence_penalty=confidence_penalty
    )
    assert torch.isclose(loss(preds, target), output)


@pytest.mark.parametrize("preds", [torch.tensor([[0.5, 0.5], [0.2, 0.8], [0.6, 0.4]])])
@pytest.mark.parametrize("target", [torch.tensor([1, 1, 0])])
@pytest.mark.parametrize("output", [torch.tensor(0.5763)])
def test_softmax_cross_entropy_loss(preds: torch.Tensor, target: torch.Tensor, output: torch.Tensor):
    loss = loss_modules.SoftmaxCrossEntropyLoss()
    assert torch.isclose(loss(preds, target), output, rtol=0.0001)


@pytest.mark.parametrize("preds", [torch.arange(6).reshape(3, 2).float()])
@pytest.mark.parametrize("target", [torch.arange(6, 12).reshape(3, 2).float()])
@pytest.mark.parametrize("output", [torch.tensor(-21.4655).float()])
def test_sigmoid_cross_entropy_loss(preds: torch.Tensor, target: torch.Tensor, output: torch.Tensor):
    loss = loss_modules.SigmoidCrossEntropyLoss()
    assert torch.isclose(loss(preds, target), output)
